"""add is_hash_migrated column

Revision ID: 784f145ec892
Revises: eef4477c37d0
Create Date: 2024-09-03 21:04:21.880497

"""

import sqlalchemy as sa
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
revision = "784f145ec892"
down_revision = "eef4477c37d0"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "currentprivacypreferencev2",
        sa.Column("is_hash_migrated", sa.Boolean(), server_default="f", nullable=False),
    )
    op.add_column(
        "custom_privacy_request_field",
        sa.Column("is_hash_migrated", sa.Boolean(), server_default="f", nullable=False),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("is_hash_migrated", sa.Boolean(), server_default="f", nullable=False),
    )
    op.add_column(
        "providedidentity",
        sa.Column("is_hash_migrated", sa.Boolean(), server_default="f", nullable=False),
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("is_hash_migrated", sa.Boolean(), server_default="f", nullable=False),
    )

    connection = op.get_bind()

    # only run the index creation if the tables have less than 1 million rows
    currentprivacypreferencev2_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM currentprivacypreferencev2")
    ).scalar()
    if currentprivacypreferencev2_count < 1000000:
        op.create_index(
            "idx_currentprivacypreferencev2_unmigrated",
            "currentprivacypreferencev2",
            ["is_hash_migrated"],
            unique=False,
            postgresql_where=sa.text("is_hash_migrated IS false"),
        )
    else:
        logger.warning(
            "The currentprivacypreferencev2 table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY idx_currentprivacypreferencev2_unmigrated "
            "ON currentprivacypreferencev2 (is_hash_migrated) "
            "WHERE is_hash_migrated IS false'"
        )

    custom_privacy_request_field_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM custom_privacy_request_field")
    ).scalar()
    if custom_privacy_request_field_count < 1000000:
        op.create_index(
            "idx_custom_privacy_request_field_unmigrated",
            "custom_privacy_request_field",
            ["is_hash_migrated"],
            unique=False,
            postgresql_where=sa.text("is_hash_migrated IS false"),
        )
    else:
        logger.warning(
            "The custom_privacy_request_field table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY idx_custom_privacy_request_field_unmigrated "
            "ON custom_privacy_request_field (is_hash_migrated) "
            "WHERE is_hash_migrated IS false'"
        )

    privacypreferencehistory_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM privacypreferencehistory")
    ).scalar()
    if privacypreferencehistory_count < 1000000:
        op.create_index(
            "idx_privacypreferencehistory_unmigrated",
            "privacypreferencehistory",
            ["is_hash_migrated"],
            unique=False,
            postgresql_where=sa.text("is_hash_migrated IS false"),
        )
    else:
        logger.warning(
            "The privacypreferencehistory table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY idx_privacypreferencehistory_unmigrated "
            "ON privacypreferencehistory (is_hash_migrated) "
            "WHERE is_hash_migrated IS false'"
        )

    providedidentity_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM providedidentity")
    ).scalar()
    if providedidentity_count < 1000000:
        op.create_index(
            "idx_providedidentity_unmigrated",
            "providedidentity",
            ["is_hash_migrated"],
            unique=False,
            postgresql_where=sa.text("is_hash_migrated IS false"),
        )
    else:
        logger.warning(
            "The providedidentity table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY idx_providedidentity_unmigrated "
            "ON providedidentity (is_hash_migrated) "
            "WHERE is_hash_migrated IS false'"
        )

    servednoticehistory_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM servednoticehistory")
    ).scalar()
    if servednoticehistory_count < 1000000:
        op.create_index(
            "idx_servednoticehistory_unmigrated",
            "servednoticehistory",
            ["is_hash_migrated"],
            unique=False,
            postgresql_where=sa.text("is_hash_migrated IS false"),
        )
    else:
        logger.warning(
            "The servednoticehistory table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY idx_servednoticehistory_unmigrated "
            "ON servednoticehistory (is_hash_migrated) "
            "WHERE is_hash_migrated IS false'"
        )


def downgrade():
    op.drop_index(
        "idx_servednoticehistory_unmigrated",
        table_name="servednoticehistory",
        postgresql_where=sa.text("is_hash_migrated IS false"),
    )
    op.drop_column("servednoticehistory", "is_hash_migrated")
    op.drop_index(
        "idx_providedidentity_unmigrated",
        table_name="providedidentity",
        postgresql_where=sa.text("is_hash_migrated IS false"),
    )
    op.drop_column("providedidentity", "is_hash_migrated")
    op.drop_index(
        "idx_privacypreferencehistory_unmigrated",
        table_name="privacypreferencehistory",
        postgresql_where=sa.text("is_hash_migrated IS false"),
    )
    op.drop_column("privacypreferencehistory", "is_hash_migrated")
    op.drop_index(
        "idx_custom_privacy_request_field_unmigrated",
        table_name="custom_privacy_request_field",
        postgresql_where=sa.text("is_hash_migrated IS false"),
    )
    op.drop_column("custom_privacy_request_field", "is_hash_migrated")
    op.drop_index(
        "idx_currentprivacypreferencev2_unmigrated",
        table_name="currentprivacypreferencev2",
        postgresql_where=sa.text("is_hash_migrated IS false"),
    )
    op.drop_column("currentprivacypreferencev2", "is_hash_migrated")



import asyncio
from functools import wraps
from typing import Any, Callable


def sync(func: Callable) -> Any:
    """Converts an async function into a sync function."""

    @wraps(func)
    def wrap(*args: Any, **kwargs: Any) -> Any:
        try:
            loop = asyncio.get_running_loop()
            return loop.run_until_complete(func(*args, **kwargs))
        except RuntimeError:
            return asyncio.run(func(*args, **kwargs))

    return wrap



from base64 import b64encode

from requests import PreparedRequest

from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.schemas.saas.strategy_configuration import (
    BasicAuthenticationConfiguration,
)
from fides.api.service.authentication.authentication_strategy import (
    AuthenticationStrategy,
)
from fides.api.util.saas_util import assign_placeholders


class BasicAuthenticationStrategy(AuthenticationStrategy):
    """
    Replaces the username and password placeholders with the actual credentials
    and uses them to add a basic authentication header to the incoming request.
    """

    name = "basic"
    configuration_model = BasicAuthenticationConfiguration

    def __init__(self, configuration: BasicAuthenticationConfiguration):
        self.username = configuration.username
        self.password = configuration.password

    def add_authentication(
        self, request: PreparedRequest, connection_config: ConnectionConfig
    ) -> PreparedRequest:
        """Add basic authentication to the request"""
        secrets = connection_config.secrets
        username = assign_placeholders(self.username, secrets)  # type: ignore
        password = assign_placeholders(self.password, secrets)  # type: ignore

        # the requests library treats a None password as a "None" literal string
        # so we need to override this behavior
        if password is None:
            password = ""

        request.headers["Authorization"] = (
            "Basic " + b64encode(f"{username}:{password}".encode()).decode()
        )
        return request



"""adds additional config vals on experience config

Revision ID: fc82ab64bd5e
Revises: a6d9cdfcc7dc
Create Date: 2024-07-29 14:55:01.609732

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fc82ab64bd5e"
down_revision = "a6d9cdfcc7dc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "experienceconfigtemplate",
        sa.Column("show_layer1_notices", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "experienceconfigtemplate",
        sa.Column(
            "layer1_button_options",
            sa.String(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("show_layer1_notices", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "layer1_button_options",
            sa.String(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("show_layer1_notices", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "layer1_button_options",
            sa.String(),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyexperienceconfighistory", "layer1_button_options")
    op.drop_column("privacyexperienceconfighistory", "show_layer1_notices")
    op.drop_column("privacyexperienceconfig", "layer1_button_options")
    op.drop_column("privacyexperienceconfig", "show_layer1_notices")
    op.drop_column("experienceconfigtemplate", "layer1_button_options")
    op.drop_column("experienceconfigtemplate", "show_layer1_notices")
    # ### end Alembic commands ###



from functools import reduce
from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, Union

import immutables
from ordered_set import OrderedSet

T = TypeVar("T")
U = TypeVar("U")

NodeInput = Dict[str, List[Any]]  # Of format {node_address: []}
Row = Dict[str, Any]
FIDESOPS_DO_NOT_MASK_INDEX = "FIDESOPS_DO_NOT_MASK"


def make_immutable(obj: Any) -> Any:
    """
    Recursively converts a mutable object into an immutable version.
    Dictionaries are converted to immutable `Map`s from the `immutables` library,
    lists are converted to tuples, and other objects are returned unchanged.
    """
    if isinstance(obj, dict):
        return immutables.Map(
            {key: make_immutable(value) for key, value in obj.items()}
        )
    if isinstance(obj, list):
        return tuple(make_immutable(item) for item in obj)
    return obj


def make_mutable(obj: Any) -> Any:
    """
    Recursively converts an immutable object into a mutable version.
    `Map`s from the `immutables` library and dictionaries are converted to mutable dictionaries,
    tuples and `OrderedSet`s are converted to lists, and other objects are returned unchanged.
    """
    if isinstance(obj, (dict, immutables.Map)):
        return {key: make_mutable(value) for key, value in obj.items()}
    if isinstance(obj, (tuple, OrderedSet)):
        return [make_mutable(item) for item in obj]
    return obj


def merge_dicts(*dicts: Dict[T, U]) -> Dict[T, U]:
    """Merge any number of dictionaries.

    Right-hand values take precedence. That is,
    merge_dicts({"A": 1, "B": 2}, {"A": 2, "C": 4})
    =>  {'A': 2, 'B': 2, 'C': 4}
    """
    if dicts:
        return reduce(lambda x, y: {**x, **y}, dicts) or {}
    return {}


def append(d: Dict[T, List[U]], key: T, value: U) -> None:
    """Append to values stored under a dictionary key.

    append({},"A",1) sets dict to {"A":[1]}
    append({"A":[1],"A",2) sets dict to {"A":[1,2]}
    append({"A":[1],"A",[2, 3, 4]) sets dict to {"A":[1, 2, 3, 4]}
    """
    if value:
        if key in d:
            if isinstance(value, list):
                d[key].extend(value)
            else:
                d[key].append(value)
        else:
            d[key] = value if isinstance(value, list) else [value]


def append_unique(d: Dict[T, OrderedSet[U]], key: T, value: Union[U, List[U]]) -> None:
    """Append to values stored under a dictionary key.

    append_unique({}, "A", 1) sets dict to {"A": {1}}
    append_unique({"A": {1}}, "A", 2) sets dict to {"A": {1, 2}}
    append_unique({"A": {1}}, "A", [2, 3, 4]) sets dict to {"A": {1, 2, 3, 4}}
    append_unique({"A": {1, 2}}, "A", [2, 3, 4]) sets dict to {"A": {1, 2, 3, 4}}
    """
    if value:
        if key not in d:
            d[key] = OrderedSet()

        if isinstance(value, list):
            d[key].update(make_immutable(value))
        else:
            d[key].add(make_immutable(value))


def partition(_iterable: Iterable[T], extractor: Callable[[T], U]) -> Dict[U, List[T]]:
    """partition a collection by the output of an arbitrary extractor function"""
    out: Dict[U, List[T]] = {}
    for t in _iterable:
        append(out, extractor(t), t)
    return out


def filter_nonempty_values(d: Optional[Dict[Any, Any]]) -> Dict[Any, Any]:
    """Return the input map with empty values removed. On an input of None
    will return an empty Dict."""
    if d:
        return {e[0]: e[1] for e in d.items() if e[1]}
    return {}


def extract_key_for_address(
    full_request_id: str, number_of_leading_strings_to_exclude: int
) -> str:
    """
    Handles extracting the correct Dataset:Collection to map to extracted
    values.

    Due to differences in the number of leading strings based on access or
    erasure, a parameter is used to ensure the correct values are returned.

    Handles an edge case where double underscores exist in either the fides_key
    of the Dataset or the Collection name.
    """
    request_id_dataset, collection = full_request_id.split(":")
    dataset = request_id_dataset.split("__", number_of_leading_strings_to_exclude)[-1]
    return f"{dataset}:{collection}"



"""This module handles finding and parsing fides configuration files."""

# pylint: disable=C0115,C0116, E0213
from typing import List, Optional, Pattern, Tuple, Union

import validators
from pydantic import Field, SerializeAsAny, ValidationInfo, field_validator
from pydantic_settings import SettingsConfigDict
from slowapi.wrappers import parse_many  # type: ignore

from fides.api.cryptography.cryptographic_util import (
    generate_salt,
    hash_credential_with_salt,
)
from fides.api.custom_types import URLOriginString
from fides.api.oauth.roles import OWNER
from fides.common.api.scope_registry import SCOPE_REGISTRY

from .fides_settings import FidesSettings

ENV_PREFIX = "FIDES__SECURITY__"


class SecuritySettings(FidesSettings):
    """Configuration settings for application security."""

    aes_encryption_key_length: int = Field(
        default=16,
        description="Length of desired encryption key when using Fides to generate a random secure string used for AES encryption.",
    )
    aes_gcm_nonce_length: int = Field(
        default=12,
        description="Length of desired random byte str for the AES GCM encryption used throughout Fides.",
    )
    app_encryption_key: str = Field(
        default="", description="The key used to sign Fides API access tokens."
    )
    cors_origins: SerializeAsAny[List[URLOriginString]] = Field(
        default_factory=list,
        description="A list of client addresses allowed to communicate with the Fides webserver.",
    )
    cors_origin_regex: Optional[Pattern] = Field(
        default=None,
        description="A regex pattern used to set the CORS origin allowlist.",
    )
    drp_jwt_secret: Optional[str] = Field(
        default=None,
        description="JWT secret by which passed-in identity is decrypted according to the HS256 algorithm.",
    )
    encoding: str = Field(
        default="UTF-8", description="Text encoding to use for the application."
    )
    env: str = Field(
        default="prod",
        description="The default, `dev`, does not apply authentication to endpoints typically used by the CLI. The other option, `prod`, requires authentication for _all_ endpoints that may contain sensitive information.",
    )
    identity_verification_attempt_limit: int = Field(
        default=3,
        description="The number of times identity verification will be attempted before raising an error.",
    )
    oauth_root_client_id: str = Field(
        default="",
        description="The value used to identify the Fides application root API client.",
    )
    oauth_root_client_secret: str = Field(
        default="",
        description="The secret value used to authenticate the Fides application root API client.",
    )
    oauth_root_client_secret_hash: Optional[Tuple] = Field(
        default=None,
        description="Automatically generated by Fides, and represents a hashed value of the oauth_root_client_secret.",
    )
    oauth_access_token_expire_minutes: int = Field(
        default=11520,
        description="The time in minutes for which Fides API tokens will be valid. Default value is equal to 8 days.",
    )
    oauth_client_id_length_bytes: int = Field(
        default=16,
        description="Sets desired length in bytes of generated client id used for oauth.",
    )
    oauth_client_secret_length_bytes: int = Field(
        default=16,
        description="Sets desired length in bytes of generated client secret used for oauth.",
    )
    parent_server_password: Optional[str] = Field(
        default=None,
        description="When using a parent/child Fides deployment, this password will be used by the child server to access the parent server.",
    )
    parent_server_username: Optional[str] = Field(
        default=None,
        description="When using a parent/child Fides deployment, this username will be used by the child server to access the parent server.",
    )
    public_request_rate_limit: str = Field(
        default="2000/minute",
        description="The number of requests from a single IP address allowed to hit a public endpoint within the specified time period",
    )
    rate_limit_prefix: str = Field(
        default="fides-",
        description="The prefix given to keys in the Redis cache used by the rate limiter.",
    )
    request_rate_limit: str = Field(
        default="1000/minute",
        description="The number of requests from a single IP address allowed to hit an endpoint within a rolling 60 second period.",
    )
    root_user_scopes: List[str] = Field(
        default=SCOPE_REGISTRY,
        description="The list of scopes that are given to the root user.",
    )
    root_user_roles: List[str] = Field(
        default=[OWNER],
        description="The list of roles that are given to the root user.",
    )
    root_password: Optional[str] = Field(
        default=None,
        description="If set, this can be used in conjunction with root_username to log in without first creating a user in the database.",
    )

    root_username: Optional[str] = Field(
        default=None,
        description="If set, this can be used in conjunction with root_password to log in without first creating a user in the database.",
    )
    subject_request_download_ui_enabled: bool = Field(
        default=False,
        description="If set to True, the user interface will display a download button for subject requests.",
    )
    dsr_testing_tools_enabled: bool = Field(
        default=False,
        description="If set to True, contributor and owner roles will be able to run test privacy requests.",
    )
    subject_request_download_link_ttl_seconds: int = Field(
        default=432000,
        description="The number of seconds that a pre-signed download URL when using S3 storage will be valid. The default is equal to 5 days.",
    )
    enable_audit_log_resource_middleware: Optional[bool] = Field(
        default=False,
        description="Either enables the collection of audit log resource data or bypasses the middleware",
    )

    bastion_server_host: Optional[str] = Field(
        default=None, description="An optional field to store the bastion server host"
    )
    bastion_server_ssh_username: Optional[str] = Field(
        default=None,
        description="An optional field to store the username used to access the bastion server",
    )
    bastion_server_ssh_private_key: Optional[str] = Field(
        default=None,
        description="An optional field to store the key used to SSH into the bastion server.",
    )
    bastion_server_ssh_timeout: float = Field(
        default=0.1,
        description="The timeout in seconds for the transport socket (``socket.settimeout``)",
    )
    bastion_server_ssh_tunnel_timeout: float = Field(
        default=10,
        description="The timeout in seconds for tunnel connection (open_channel timeout)",
    )

    @field_validator("app_encryption_key", mode="before")
    @classmethod
    def validate_encryption_key_length(
        cls, v: Optional[str], info: ValidationInfo
    ) -> Optional[str]:
        """Validate the encryption key is exactly 32 characters"""

        # If the value is the default value, return immediately to prevent unwanted errors
        if v == "":
            return v

        if v is None or len(v.encode(info.data.get("encoding", "UTF-8"))) != 32:
            raise ValueError(
                "APP_ENCRYPTION_KEY value must be exactly 32 characters long"
            )
        return v

    @field_validator("cors_origins", mode="before")
    @classmethod
    def assemble_cors_origins(cls, v: Union[str, List[str]]) -> Union[List[str], str]:
        """Return a list of valid origins for CORS requests"""

        def validate(values: List[str]) -> None:
            for value in values:
                if value != "*":
                    if not validators.url(value):
                        raise ValueError(f"{value} is not a valid url")

        if isinstance(v, str) and not v.startswith("["):
            values = [i.strip() for i in v.split(",")]
            validate(values)

            return values
        if isinstance(v, (list, str)):
            validate(v)  # type: ignore

            return v
        raise ValueError(v)

    @field_validator("oauth_root_client_secret_hash", mode="before")
    @classmethod
    def assemble_root_access_token(
        cls, v: Optional[str], info: ValidationInfo
    ) -> Optional[Tuple]:
        """
        Sets a hashed value of the root access key.
        This is hashed as it is not wise to return a plaintext for of the
        root credential anywhere in the system.
        """
        value = info.data.get("oauth_root_client_secret", "")

        if not value:
            return None

        encoding = info.data.get("encoding", "UTF-8")

        salt = generate_salt()
        hashed_client_id = hash_credential_with_salt(
            value.encode(encoding), salt.encode(encoding)
        )
        oauth_root_client_secret_hash = (hashed_client_id, salt.encode(encoding))  # type: ignore
        return oauth_root_client_secret_hash

    @field_validator("request_rate_limit")
    @classmethod
    def validate_request_rate_limit(
        cls,
        v: str,
    ) -> str:
        """Validate the formatting of `request_rate_limit`"""
        try:
            # Defer to `limits.parse_many` https://limits.readthedocs.io/en/stable/api.html#limits.parse_many
            parse_many(v)
        except ValueError:
            message = """
            Ratelimits must be specified in the format: [count] [per|/] [n (optional)] [second|minute|hour|day|month|year]
            e.g. 10 per hour
            e.g. 10/hour
            e.g. 10/hour;100/day;2000 per year
            e.g. 100/day, 500/7days
            """
            raise ValueError(message)
        return v

    @field_validator("env")
    @classmethod
    def validate_env(
        cls,
        v: str,
    ) -> str:
        """Validate the formatting of `request_rate_limit`"""
        if v not in ["dev", "prod"]:
            message = "Security environment must be either 'dev' or 'prod'."
            raise ValueError(message)
        return v

    model_config = SettingsConfigDict(env_prefix=ENV_PREFIX)



from typing import Annotated, Any, Dict, List, Optional

from fastapi import Depends, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session
from starlette.exceptions import HTTPException
from starlette.status import (
    HTTP_200_OK,
    HTTP_204_NO_CONTENT,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from fides.api.api import deps
from fides.api.common_exceptions import (
    DataCategoryNotSupported,
    DrpActionValidationError,
    KeyOrNameAlreadyExists,
    PolicyValidationError,
    RuleTargetValidationError,
    RuleValidationError,
)
from fides.api.models.client import ClientDetail
from fides.api.models.policy import Policy, Rule, RuleTarget
from fides.api.models.sql_models import DataCategory  # type: ignore
from fides.api.models.storage import StorageConfig
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas import policy as schemas
from fides.api.schemas.api import BulkUpdateFailed
from fides.api.schemas.policy import ActionType
from fides.api.util.api_router import APIRouter
from fides.api.util.logger import Pii
from fides.common.api import scope_registry
from fides.common.api.v1 import urn_registry as urls

router = APIRouter(tags=["DSR Policy"], prefix=urls.V1_URL_PREFIX)


@router.get(
    urls.POLICY_LIST,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.PolicyResponse],
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.POLICY_READ])],
)
def get_policy_list(
    *,
    db: Session = Depends(deps.get_db),
    params: Params = Depends(),
) -> AbstractPage[Policy]:
    """
    Return a paginated list of all Policy records in this system
    """
    logger.debug("Finding all policies with pagination params '{}'", params)
    policies = Policy.query(db=db).order_by(Policy.created_at.desc())
    return paginate(policies, params=params)


def get_policy_or_error(db: Session, policy_key: FidesKey) -> Policy:
    """Helper method to load Policy or throw a 404"""
    logger.debug("Finding policy with key '{}'", policy_key)
    policy = Policy.get_by(db=db, field="key", value=policy_key)
    if not policy:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Policy found for key {policy_key}.",
        )

    return policy


@router.get(
    urls.POLICY_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.PolicyResponse,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.POLICY_READ])],
)
def get_policy(
    *,
    policy_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> schemas.Policy:
    """
    Return a single Policy
    """
    return get_policy_or_error(db, policy_key)  # type: ignore[return-value]


@router.patch(
    urls.POLICY_LIST,
    status_code=HTTP_200_OK,
    response_model=schemas.BulkPutPolicyResponse,
)
def create_or_update_policies(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[scope_registry.POLICY_CREATE_OR_UPDATE],
    ),
    db: Session = Depends(deps.get_db),
    data: Annotated[List[schemas.Policy], Field(max_length=50)],  # type: ignore
) -> schemas.BulkPutPolicyResponse:
    """
    Given a list of policy data elements, create or update corresponding Policy objects
    or report failure
    """
    created_or_updated: List[Policy] = []
    failed: List[BulkUpdateFailed] = []
    logger.info("Starting bulk upsert for {} policies", len(data))

    for policy_schema in data:
        policy_data: Dict[str, Any] = dict(policy_schema)
        try:
            policy = Policy.create_or_update(
                db=db,
                data={
                    "name": policy_data["name"],
                    "key": policy_data.get("key"),
                    "drp_action": policy_data.get("drp_action"),
                    "execution_timeframe": policy_data.get("execution_timeframe"),
                },
            )
        except (
            KeyOrNameAlreadyExists,
            DrpActionValidationError,
            IntegrityError,
        ) as exc:
            logger.warning("Create/update failed for policy: {}", Pii(str(exc)))
            failure = {
                "message": exc.args[0],
                "data": policy_data,
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except PolicyValidationError as exc:
            logger.warning("Create/update failed for policy: {}", Pii(str(exc)))
            failure = {
                "message": "This record could not be added because the data provided was invalid.",
                "data": policy_data,
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        else:
            created_or_updated.append(policy)  # type: ignore[arg-type]

    return schemas.BulkPutPolicyResponse(
        succeeded=created_or_updated,
        failed=failed,
    )


def get_rule_or_error(db: Session, policy_key: FidesKey, rule_key: FidesKey) -> Rule:
    """
    Helper method to load Rule or throw a 404 if it can't be found.
    Also throws a 404 if a `Policy` with the given key can't be found.
    """
    policy = get_policy_or_error(db, policy_key)
    logger.debug("Finding rule with key '{}'", rule_key)
    rule = Rule.filter(
        db=db,
        conditions=((Rule.policy_id == policy.id) & (Rule.key == rule_key)),
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} attached to policy {policy_key}",
        )

    return rule


@router.get(
    urls.RULE_LIST,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.RuleResponseWithTargets],
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule_list(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    params: Params = Depends(),
) -> AbstractPage[Rule]:
    """
    Return a paginated list of all `Rule` records associated with the given `Policy`.
    Throws a 404 if the given `Policy` can't be found.
    """
    policy = get_policy_or_error(db, policy_key)
    logger.debug(
        "Finding all rules for policy {} with pagination params '{}'",
        policy_key,
        params,
    )
    rules = Rule.filter(
        db=db,
        conditions=(Rule.policy_id == policy.id),
    ).order_by(Rule.created_at.desc())
    return paginate(rules, params=params)


@router.get(
    urls.RULE_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.RuleResponseWithTargets,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> Rule:
    """
    Return a single `Rule` with the given key, associated with the given `Policy`
    """
    return get_rule_or_error(db, policy_key, rule_key)


@router.patch(
    urls.RULE_LIST,
    status_code=HTTP_200_OK,
    response_model=schemas.BulkPutRuleResponse,
)
def create_or_update_rules(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[scope_registry.RULE_CREATE_OR_UPDATE],
    ),
    policy_key: FidesKey,
    db: Session = Depends(deps.get_db),
    input_data: Annotated[List[schemas.RuleCreate], Field(max_length=50)],  # type: ignore
) -> schemas.BulkPutRuleResponse:
    """
    Given a list of Rule data elements, create or update corresponding Rule objects
    or report failure
    """
    logger.info("Finding policy with key '{}'", policy_key)

    policy = get_policy_or_error(db, policy_key)

    created_or_updated: List[Rule] = []
    failed: List[BulkUpdateFailed] = []

    logger.info(
        "Starting bulk upsert for {} rules on policy {}", len(input_data), policy_key
    )

    for schema in input_data:
        # Validate all FKs in the input data exist
        associated_storage_config_id = None
        if schema.action_type == ActionType.access:
            # Only validate the associated StorageConfig on access rules
            storage_destination_key = schema.storage_destination_key
            # storage key doesn't need to be specified, as there is a default to fallback to
            if storage_destination_key:
                associated_storage_config: Optional[StorageConfig] = (
                    StorageConfig.get_by(
                        db=db,
                        field="key",
                        value=storage_destination_key,
                    )
                )
                if not associated_storage_config:
                    logger.warning(
                        "No storage config found with key {}", storage_destination_key
                    )
                    failure = {
                        "message": f"A StorageConfig with key {storage_destination_key} does not exist",
                        "data": dict(
                            schema
                        ),  # Be sure to pass the schema out the same way it came in
                    }
                    failed.append(BulkUpdateFailed(**failure))
                    continue
                associated_storage_config_id = associated_storage_config.id

        masking_strategy_data = None
        if schema.masking_strategy:
            masking_strategy_data = schema.masking_strategy.model_dump(mode="json")

        try:
            rule = Rule.create_or_update(
                db=db,
                data={
                    "action_type": schema.action_type,
                    "key": schema.key,
                    "name": schema.name,
                    "policy_id": policy.id,
                    "storage_destination_id": associated_storage_config_id,
                    "masking_strategy": masking_strategy_data,
                },
            )
        except KeyOrNameAlreadyExists as exc:
            logger.warning(
                "Create/update failed for rule '{}' on policy {}: {}",
                schema.key,
                policy_key,
                exc,
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except RuleValidationError as exc:
            logger.warning(
                "Create/update failed for rule '{}' on policy {}: {}",
                schema.key,
                policy_key,
                Pii(str(exc)),
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except ValueError as exc:
            logger.warning(
                "Create/update failed for rule '{}' on policy {}: {}",
                schema.key,
                policy_key,
                Pii(str(exc)),
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        else:
            created_or_updated.append(rule)  # type: ignore[arg-type]

    return schemas.BulkPutRuleResponse(succeeded=created_or_updated, failed=failed)


@router.delete(
    urls.RULE_DETAIL,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_DELETE])],
)
def delete_rule(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> None:
    """
    Delete a policy rule.
    """
    policy = get_policy_or_error(db, policy_key)

    logger.info("Finding rule with key '{}'", rule_key)

    rule = Rule.filter(
        db=db, conditions=(Rule.key == rule_key and Rule.policy_id == policy.id)  # type: ignore[arg-type]
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} on Policy {policy_key}.",
        )

    logger.info("Deleting rule with key '{}'", rule_key)
    rule.delete(db=db)


def get_rule_target_or_error(
    db: Session,
    policy_key: FidesKey,
    rule_key: FidesKey,
    rule_target_key: FidesKey,
) -> RuleTarget:
    """
    Helper method to load Rule Target or throw a 404.
    Also throws a 404 if a `Policy` or `Rule` with the given keys can't be found.
    """
    logger.debug("Finding rule target with key '{}'", rule_target_key)
    rule: Rule = get_rule_or_error(db, policy_key, rule_key)
    rule_target = RuleTarget.filter(
        db=db,
        conditions=(
            (RuleTarget.rule_id == rule.id) & (RuleTarget.key == rule_target_key)
        ),
    ).first()
    if not rule_target:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule Target found for key {rule_target_key} attached to rule {rule_key}.",
        )

    return rule_target


@router.get(
    urls.RULE_TARGET_LIST,
    status_code=HTTP_200_OK,
    response_model=Page[schemas.RuleTarget],
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule_target_list(
    *,
    db: Session = Depends(deps.get_db),
    policy_key: FidesKey,
    rule_key: FidesKey,
    params: Params = Depends(),
) -> AbstractPage[RuleTarget]:
    """
    Return a paginated list of all `Rule Target` records associated with the given `Rule`
    Throws a 404 if the given `Rule` or `Policy` can't be found.
    """
    rule = get_rule_or_error(db, policy_key, rule_key)
    logger.debug(
        "Finding all rule targets for rule {} with pagination params '{}'",
        rule_key,
        params,
    )
    rule_targets = RuleTarget.filter(
        db=db,
        conditions=(RuleTarget.rule_id == rule.id),
    ).order_by(RuleTarget.created_at.desc())
    return paginate(rule_targets, params=params)


@router.get(
    urls.RULE_TARGET_DETAIL,
    status_code=HTTP_200_OK,
    response_model=schemas.RuleTarget,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_READ])],
)
def get_rule_target(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    rule_target_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> RuleTarget:
    """
    Return a single `Rule Target` associated with the given `Rule` and `Policy`
    """
    return get_rule_target_or_error(db, policy_key, rule_key, rule_target_key)


def _validate_data_categories(
    db: Session,
    data_categories: List[str],
) -> None:
    """
    Helper method to validate that all data categories supplied to the endpoint
    exist in the database
    """
    from_db = [
        dc.fides_key
        for dc in DataCategory.filter(
            db=db,
            conditions=DataCategory.fides_key.in_(data_categories),
        ).all()
    ]
    invalid_categories = list(set(data_categories) - set(from_db))
    if invalid_categories:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Invalid data categories: {invalid_categories}",
        )


@router.patch(
    urls.RULE_TARGET_LIST,
    status_code=HTTP_200_OK,
    response_model=schemas.BulkPutRuleTargetResponse,
)
def create_or_update_rule_targets(
    *,
    client: ClientDetail = Security(
        verify_oauth_client, scopes=[scope_registry.RULE_CREATE_OR_UPDATE]
    ),
    policy_key: FidesKey,
    rule_key: FidesKey,
    db: Session = Depends(deps.get_db),
    input_data: Annotated[List[schemas.RuleTarget], Field(max_length=50)],  # type: ignore
) -> schemas.BulkPutRuleTargetResponse:
    """
    Given a list of Rule data elements, create corresponding Rule objects
    or report failure
    """
    policy = get_policy_or_error(db, policy_key)

    logger.info("Finding rule with key '{}'", rule_key)
    rule = Rule.filter(
        db=db, conditions=(Rule.key == rule_key and Rule.policy_id == policy.id)  # type: ignore[arg-type]
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} on Policy {policy_key}.",
        )

    _validate_data_categories(
        db=db,
        data_categories=[schema.data_category for schema in input_data],
    )

    created_or_updated = []
    failed = []
    logger.info(
        "Starting bulk upsert for {} rule targets on rule {}", len(input_data), rule_key
    )
    for schema in input_data:
        try:
            target = RuleTarget.create_or_update(
                db=db,
                data={
                    "name": schema.name,
                    "key": schema.key,
                    "data_category": schema.data_category,
                    "rule_id": rule.id,
                },
            )
        except KeyOrNameAlreadyExists as exc:
            logger.warning(
                "Create/update failed for rule target {} on rule {}: {}",
                schema.key,
                rule_key,
                exc,
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except (
            DataCategoryNotSupported,
            PolicyValidationError,
            RuleTargetValidationError,
        ) as exc:
            logger.warning(
                "Create/update failed for rule target {} on rule {}: {}",
                schema.key,
                rule_key,
                Pii(str(exc)),
            )
            failure = {
                "message": exc.args[0],
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except IntegrityError as exc:
            logger.warning(
                "Create/update failed for rule target {} on rule {}: {}",
                schema.key,
                rule_key,
                Pii(str(exc)),
            )
            failure = {
                "message": f"DataCategory {schema.data_category} is already specified on Rule with ID {rule.id}",
                "data": dict(schema),
            }
            failed.append(BulkUpdateFailed(**failure))
        else:
            created_or_updated.append(target)

    return schemas.BulkPutRuleTargetResponse(
        succeeded=created_or_updated,
        failed=failed,
    )


@router.delete(
    urls.RULE_TARGET_DETAIL,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[scope_registry.RULE_DELETE])],
)
def delete_rule_target(
    *,
    policy_key: FidesKey,
    rule_key: FidesKey,
    rule_target_key: FidesKey,
    db: Session = Depends(deps.get_db),
) -> None:
    """
    Delete the rule target.
    """
    policy = get_policy_or_error(db, policy_key)

    logger.info("Finding rule with key '{}'", rule_key)
    rule = Rule.filter(
        db=db, conditions=(Rule.key == rule_key and Rule.policy_id == policy.id)  # type: ignore[arg-type]
    ).first()
    if not rule:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No Rule found for key {rule_key} on Policy {policy_key}.",
        )

    logger.info("Finding rule target with key '{}'", rule_target_key)
    target = RuleTarget.filter(
        db=db,
        conditions=(
            RuleTarget.key == rule_target_key and RuleTarget.rule_id == rule.id  # type: ignore[arg-type]
        ),
    ).first()
    if not target:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No RuleTarget found for key {rule_target_key} at Rule {rule_key} on Policy {policy_key}.",
        )

    logger.info("Deleting rule target with key '{}'", rule_target_key)

    target.delete(db=db)



from typing import Annotated, Dict, List, Optional

from fastapi import Body, Depends, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import Field
from requests import RequestException
from sqlalchemy.orm import Session
from starlette.exceptions import HTTPException
from starlette.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from fides.api.api import deps
from fides.api.common_exceptions import KeyOrNameAlreadyExists, StorageUploadError
from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.models.storage import (
    StorageConfig,
    default_storage_config_name,
    get_active_default_storage_config,
    get_default_storage_config_by_type,
)
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.api import BulkUpdateFailed
from fides.api.schemas.connection_configuration.connection_secrets import (
    TestStatusMessage,
)
from fides.api.schemas.storage.data_upload_location_response import DataUpload
from fides.api.schemas.storage.storage import (
    FULLY_CONFIGURED_STORAGE_TYPES,
    AWSAuthMethod,
    BulkPutStorageConfigResponse,
    StorageConfigStatus,
    StorageConfigStatusMessage,
    StorageDestination,
    StorageDestinationBase,
    StorageDestinationResponse,
    StorageDetails,
    StorageType,
)
from fides.api.schemas.storage.storage_secrets_docs_only import possible_storage_secrets
from fides.api.service.storage.storage_authenticator_service import secrets_are_valid
from fides.api.service.storage.storage_uploader_service import upload
from fides.api.util.api_router import APIRouter
from fides.api.util.logger import Pii
from fides.api.util.storage_util import get_schema_for_secrets
from fides.common.api.scope_registry import (
    STORAGE_CREATE_OR_UPDATE,
    STORAGE_DELETE,
    STORAGE_READ,
)
from fides.common.api.v1.urn_registry import (
    STORAGE_ACTIVE_DEFAULT,
    STORAGE_BY_KEY,
    STORAGE_CONFIG,
    STORAGE_DEFAULT,
    STORAGE_DEFAULT_BY_TYPE,
    STORAGE_DEFAULT_SECRETS,
    STORAGE_SECRETS,
    STORAGE_STATUS,
    STORAGE_UPLOAD,
    V1_URL_PREFIX,
)

router = APIRouter(tags=["Storage"], prefix=V1_URL_PREFIX)


@router.post(
    STORAGE_UPLOAD,
    status_code=HTTP_201_CREATED,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_CREATE_OR_UPDATE])],
    response_model=DataUpload,
)
def upload_data(
    request_id: str,
    *,
    db: Session = Depends(deps.get_db),
    data: Dict = Body(...),
    storage_key: FidesKey = Body(...),
) -> DataUpload:
    """
    Uploads data from an access request to specified storage destination.
    Returns location of data.
    """
    logger.info("Finding privacy request with id '{}'", request_id)

    privacy_request = PrivacyRequest.get(db, object_id=request_id)
    if not privacy_request:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No privacy with id {request_id}.",
        )

    logger.info("Starting storage upload for request id: {}", request_id)
    try:
        data_location: str = upload(
            db, privacy_request=privacy_request, data=data, storage_key=storage_key
        )
    except StorageUploadError as e:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail=str(e))
    except RequestException as e:
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=str(e))
    return DataUpload(location=data_location)


@router.patch(
    STORAGE_CONFIG,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_CREATE_OR_UPDATE])],
    response_model=BulkPutStorageConfigResponse,
)
def patch_config(
    *,
    db: Session = Depends(deps.get_db),
    storage_configs: Annotated[List[StorageDestination], Field(max_length=50)],  # type: ignore
) -> BulkPutStorageConfigResponse:
    """
    Given a list of storage destination elements, create or update corresponding StorageConfig objects
    or report failure.
    """
    created_or_updated: List[StorageConfig] = []
    failed: List[BulkUpdateFailed] = []

    logger.info("Starting bulk upsert for {} storage configs", len(storage_configs))
    for destination in storage_configs:
        try:
            storage_config = StorageConfig.create_or_update(
                db=db, data=destination.model_dump(mode="json")
            )
        except KeyOrNameAlreadyExists as exc:
            logger.warning(
                "Create/update failed for storage config {}: {}",
                destination.key,
                exc,
            )
            failure = {
                "message": exc.args[0],
                "data": destination.model_dump(mode="json"),
            }
            failed.append(BulkUpdateFailed(**failure))
            continue
        except Exception as exc:
            logger.warning(
                "Create/update failed for storage config {}: {}",
                destination.key,
                Pii(str(exc)),
            )
            failed.append(
                BulkUpdateFailed(
                    **{
                        "message": "Error creating or updating storage config.",
                        "data": destination.model_dump(mode="json"),
                    }
                )
            )
        else:
            created_or_updated.append(storage_config)

    return BulkPutStorageConfigResponse(succeeded=created_or_updated, failed=failed)


@router.put(
    STORAGE_SECRETS,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_CREATE_OR_UPDATE])],
    response_model=TestStatusMessage,
)
def put_config_secrets(
    config_key: FidesKey,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_storage_secrets: possible_storage_secrets,
    verify: Optional[bool] = True,
) -> TestStatusMessage:
    """
    Add or update secrets for storage config.
    """
    logger.info("Finding storage config with key '{}'", config_key)
    storage_config = StorageConfig.get_by(db=db, field="key", value=config_key)
    if not storage_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No storage configuration with key {config_key}.",
        )
    try:
        secrets_schema = get_schema_for_secrets(
            storage_type=storage_config.type,
            secrets=unvalidated_storage_secrets,
        )
    except KeyError as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=exc.args[0],
        )
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    logger.info("Updating storage config secrets for config with key '{}'", config_key)
    try:
        storage_config.set_secrets(db=db, storage_secrets=secrets_schema.model_dump(mode="json"))  # type: ignore
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    msg = f"Secrets updated for StorageConfig with key: {config_key}."
    if verify:
        status = secrets_are_valid(secrets_schema, storage_config.type)
        if status:
            logger.info(
                "Storage secrets are valid for config with key '{}'", config_key
            )
        else:
            logger.warning(
                "Storage secrets are invalid for config with key '{}'", config_key
            )

        return TestStatusMessage(
            msg=msg,
            test_status=(
                ConnectionTestStatus.succeeded
                if status
                else ConnectionTestStatus.failed
            ),
        )

    return TestStatusMessage(msg=msg, test_status=None)


@router.get(
    STORAGE_CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_READ])],
    response_model=Page[StorageDestinationResponse],
)
def get_configs(
    *, db: Session = Depends(deps.get_db), params: Params = Depends()
) -> AbstractPage[StorageConfig]:
    """
    Retrieves configs for storage.
    """
    logger.debug("Finding all storage configurations with pagination params {}", params)
    return paginate(
        StorageConfig.query(db).order_by(StorageConfig.created_at.desc()), params=params
    )


@router.get(
    STORAGE_BY_KEY,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_READ])],
    response_model=StorageDestinationResponse,
)
def get_config_by_key(
    config_key: FidesKey, *, db: Session = Depends(deps.get_db)
) -> Optional[StorageConfig]:
    """
    Retrieves configs for storage by key.
    """
    logger.debug("Finding storage config with key '{}'", config_key)

    storage_config = StorageConfig.get_by(db, field="key", value=config_key)
    if not storage_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No configuration with key {config_key}.",
        )
    return storage_config


@router.delete(
    STORAGE_BY_KEY,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_DELETE])],
)
def delete_config_by_key(
    config_key: FidesKey, *, db: Session = Depends(deps.get_db)
) -> None:
    """
    Deletes configs by key.
    """
    logger.info("Finding storage config with key '{}'", config_key)

    storage_config: Optional[StorageConfig] = StorageConfig.get_by(
        db, field="key", value=config_key
    )
    if not storage_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No configuration with key {config_key}.",
        )
    if storage_config.is_default:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Default storage configurations cannot be deleted.",
        )

    logger.info("Deleting storage config with key '{}'", config_key)
    storage_config.delete(db)


# this needs to come before other `/default/{storage_type}` routes so that `/status`
# isn't picked up as a path param
@router.get(
    STORAGE_ACTIVE_DEFAULT,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_READ])],
    response_model=StorageDestinationResponse,
)
def get_active_default_config(
    *, db: Session = Depends(deps.get_db)
) -> Optional[StorageConfig]:
    """
    Retrieves the active default storage config.
    """
    logger.debug("Finding active default storage config")
    storage_config = get_active_default_storage_config(db)
    if not storage_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No active default storage config found.",
        )
    return storage_config


# this needs to come before other `/default/{storage_type}` routes so that `/status`
# isn't picked up as a path param
@router.get(
    STORAGE_STATUS,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_READ])],
    response_model=StorageConfigStatusMessage,
    responses={
        HTTP_200_OK: {
            "content": {
                "application/json": {
                    "example": {
                        "config_status": "configured",
                        "detail": "Active default storage location of type s3 is fully configured",
                    }
                }
            }
        }
    },
)
def get_storage_status(
    *, db: Session = Depends(deps.get_db)
) -> StorageConfigStatusMessage:
    """
    Determines the status of the active default storage config.

    In order to be considered fully configured, the active default storage
    config MUST be an s3 storage config.
    """
    logger.info("Determining active default storage config status")

    # confirm an active default storage config is present
    storage_config = get_active_default_storage_config(db)
    if not storage_config:
        return StorageConfigStatusMessage(
            config_status=StorageConfigStatus.not_configured,
            detail="No active default storage configuration found",
        )

    # at this point, we only treat the active default storage as fully configured
    # if it is an s3 storage location. eventually we may need to adjust this when
    # we support other object stores/storage destinations as valid for production.
    if storage_config.type not in FULLY_CONFIGURED_STORAGE_TYPES:
        return StorageConfigStatusMessage(
            config_status=StorageConfigStatus.not_configured,
            detail="Active default storage configuration is not one of the fully configured storage types",
        )

    try:
        details = storage_config.details
        StorageDestinationBase.validate_details(details, storage_config.type.value)  # type: ignore
    except Exception as e:
        logger.error(f"Invalid or unpopulated details on {storage_config.type.value} storage configuration: {Pii(str(e))}")  # type: ignore
        return StorageConfigStatusMessage(
            config_status=StorageConfigStatus.not_configured,
            detail=f"Invalid or unpopulated details on {storage_config.type.value} storage configuration",  # type: ignore
        )

    # currently, secrets are only required for s3 storage with auth method `SECRET_KEYS`
    secrets = storage_config.secrets
    if _storage_config_requires_secrets(storage_config):
        if not secrets:
            return StorageConfigStatusMessage(
                config_status=StorageConfigStatus.not_configured,
                detail=f"No secrets found for {storage_config.type.value} storage configuration",  # type: ignore
            )
        try:
            get_schema_for_secrets(
                storage_type=storage_config.type,  # type: ignore
                secrets=secrets,
            )
        except (ValueError, KeyError) as e:
            logger.error(f"Invalid secrets found on {storage_config.type.value} storage configuration: {Pii(str(e))}")  # type: ignore
            return StorageConfigStatusMessage(
                config_status=StorageConfigStatus.not_configured,
                detail=f"Invalid secrets found on {storage_config.type.value} storage configuration",  # type: ignore
            )

    return StorageConfigStatusMessage(
        config_status=StorageConfigStatus.configured,
        detail=f"Active default storage configuration of type {storage_config.type.value} is fully configured",  # type: ignore
    )


def _storage_config_requires_secrets(storage_config: StorageConfig) -> bool:
    return (
        storage_config.details.get(StorageDetails.AUTH_METHOD.value, None)
        == AWSAuthMethod.SECRET_KEYS.value
    )


@router.put(
    STORAGE_DEFAULT,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_CREATE_OR_UPDATE])],
    response_model=StorageDestinationResponse,
)
def put_default_config(
    *,
    db: Session = Depends(deps.get_db),
    incoming_storage_config: StorageDestinationBase,
) -> StorageConfig:
    """
    Create or update the default storage configuration for the given storage type
    """
    logger.info(
        "Starting upsert for default storage of type '{}'", incoming_storage_config.type
    )

    incoming_data = incoming_storage_config.model_dump(mode="json")
    existing_default = get_default_storage_config_by_type(
        db, incoming_storage_config.type
    )
    if existing_default:
        # take the key of the existing default and add that to the incoming data, to ensure we overwrite the same record
        incoming_data["key"] = existing_default.key
    else:
        # set a name for our config if we're creating a new default
        incoming_data["name"] = default_storage_config_name(
            incoming_storage_config.type  # type: ignore
        )

    # since we're setting the default storage config for the given type, `is_default` MUST be set to `True`
    incoming_data["is_default"] = True

    try:
        storage_config = StorageConfig.create_or_update(db=db, data=incoming_data)
        return storage_config
    except KeyOrNameAlreadyExists as exc:
        logger.warning(
            "Create/update failed for default config update for storage type {}: {}",
            incoming_storage_config.type,
            exc,
        )
        raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=exc.args[0])
    except Exception as exc:
        logger.warning(
            "Create/update failed for default config update for storage type {}: {}",
            incoming_storage_config.type,
            Pii(str(exc)),
        )
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Error creating or updating storage config.",
        )


@router.put(
    STORAGE_DEFAULT_SECRETS,
    status_code=HTTP_200_OK,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_CREATE_OR_UPDATE])],
    response_model=TestStatusMessage,
)
def put_default_config_secrets(
    storage_type: StorageType,
    *,
    db: Session = Depends(deps.get_db),
    unvalidated_storage_secrets: possible_storage_secrets,
    verify: Optional[bool] = True,
) -> TestStatusMessage:
    """
    Add or update secrets for the default storage config of the given type
    """
    logger.info("Finding default config of storage type '{}'", storage_type.value)
    storage_config = get_default_storage_config_by_type(db, storage_type)
    if not storage_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No default configuration found for storage type {storage_type.value}.",
        )
    try:
        secrets_schema = get_schema_for_secrets(
            storage_type=storage_config.type,
            secrets=unvalidated_storage_secrets,
        )
    except KeyError as exc:
        raise HTTPException(
            status_code=HTTP_422_UNPROCESSABLE_ENTITY,
            detail=exc.args[0],
        )
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    logger.info(
        "Updating storage config secrets for default config of storage type '{}'",
        storage_type.value,
    )
    try:
        storage_config.set_secrets(db=db, storage_secrets=secrets_schema.model_dump(mode="json"))  # type: ignore
    except ValueError as exc:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=exc.args[0],
        )

    msg = f"Secrets updated for default config of storage type: {storage_type.value}."
    if verify:
        status = secrets_are_valid(secrets_schema, storage_config.type)
        if status:
            logger.info(
                "Storage secrets are valid for default config of storage type '{}'",
                storage_type.value,
            )
        else:
            logger.warning(
                "Storage secrets are invalid for default config of storage type '{}'",
                storage_type.value,
            )

        return TestStatusMessage(
            msg=msg,
            test_status=(
                ConnectionTestStatus.succeeded
                if status
                else ConnectionTestStatus.failed
            ),
        )

    return TestStatusMessage(msg=msg, test_status=None)


@router.get(
    STORAGE_DEFAULT,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_READ])],
    response_model=Page[StorageDestinationResponse],
)
def get_default_configs(
    *, db: Session = Depends(deps.get_db), params: Params = Depends()
) -> AbstractPage[StorageConfig]:
    """
    Retrieves default configs for each storage types.
    """
    logger.debug(
        "Finding default storage configurations with pagination params {}", params
    )
    return paginate(
        db.query(StorageConfig)
        .filter_by(is_default=True)
        .order_by(StorageConfig.created_at.desc()),
        params=params,
    )


@router.get(
    STORAGE_DEFAULT_BY_TYPE,
    dependencies=[Security(verify_oauth_client, scopes=[STORAGE_READ])],
    response_model=StorageDestinationResponse,
)
def get_default_config_by_type(
    storage_type: StorageType, *, db: Session = Depends(deps.get_db)
) -> Optional[StorageConfig]:
    """
    Retrieves default config for given storage type.
    """
    logger.debug("Finding default config for storage type '{}'", storage_type.value)
    storage_config = get_default_storage_config_by_type(db, storage_type)
    if not storage_config:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No default config for storage type {storage_type.value}.",
        )
    return storage_config



from typing import Optional

import rich_click as click
from click_default_group import DefaultGroup

from fides.cli.options import fides_key_argument, manifests_dir_argument
from fides.cli.utils import with_analytics, with_server_health_check
from fides.common.utils import echo_red
from fides.core import parse as _parse
from fides.core import pull as _pull
from fides.core.utils import git_is_dirty


@click.group(cls=DefaultGroup, default="all", default_if_no_args=True)  # type: ignore
@click.pass_context
def pull(ctx: click.Context) -> None:
    """
    Update local resource files based on the state of the objects on the server.
    """


@pull.command(name="all")  # type: ignore
@click.pass_context
@manifests_dir_argument
@click.option(
    "--all-resources",
    "-a",
    default=None,
    help="Pulls all locally missing resources from the server into this file.",
)
@with_analytics
@with_server_health_check
def pull_all(
    ctx: click.Context,
    manifests_dir: str,
    all_resources: Optional[str],
) -> None:
    """
    Retrieve all resources from the server and update the local manifest files.
    """

    # Make the resources that are pulled configurable
    config = ctx.obj["CONFIG"]
    # Do this to validate the manifests since they won't get parsed during the pull process
    _parse.parse(manifests_dir)
    if git_is_dirty(manifests_dir):
        echo_red(
            f"There are unstaged changes in your manifest directory: '{manifests_dir}' \nAborting pull!"
        )
        raise SystemExit(1)
    _pull.pull(
        url=config.cli.server_url,
        manifests_dir=manifests_dir,
        headers=config.user.auth_header,
        all_resources_file=all_resources,
        fides_key=None,
        resource_type=None,
    )


@pull.command(name="dataset")  # type: ignore
@click.pass_context
@fides_key_argument
@manifests_dir_argument
def dataset(
    ctx: click.Context,
    fides_key: str,
    manifests_dir: str,
) -> None:
    """
    Retrieve a specific dataset from the server and update the local manifest files.
    """

    config = ctx.obj["CONFIG"]
    _pull.pull(
        url=config.cli.server_url,
        manifests_dir=manifests_dir,
        headers=config.user.auth_header,
        fides_key=fides_key,
        resource_type="dataset",
        all_resources_file=None,
    )


@pull.command(name="system")  # type: ignore
@click.pass_context
@fides_key_argument
@manifests_dir_argument
def system(
    ctx: click.Context,
    fides_key: str,
    manifests_dir: str,
) -> None:
    """
    Retrieve a specific system from the server and update the local manifest files.
    """

    config = ctx.obj["CONFIG"]
    _pull.pull(
        url=config.cli.server_url,
        manifests_dir=manifests_dir,
        headers=config.user.auth_header,
        fides_key=fides_key,
        resource_type="system",
        all_resources_file=None,
    )


@pull.command(name="data_category")  # type: ignore
@click.pass_context
@fides_key_argument
@manifests_dir_argument
def data_category(
    ctx: click.Context,
    fides_key: str,
    manifests_dir: str,
) -> None:
    """
    Retrieve a specific data_category from the server and update the local manifest files.
    """

    config = ctx.obj["CONFIG"]
    _pull.pull(
        url=config.cli.server_url,
        manifests_dir=manifests_dir,
        headers=config.user.auth_header,
        fides_key=fides_key,
        resource_type="data_category",
        all_resources_file=None,
    )


@pull.command(name="data_use")  # type: ignore
@click.pass_context
@fides_key_argument
@manifests_dir_argument
def data_use(
    ctx: click.Context,
    fides_key: str,
    manifests_dir: str,
) -> None:
    """
    Retrieve a specific data_use from the server and update the local manifest files.
    """

    config = ctx.obj["CONFIG"]
    _pull.pull(
        url=config.cli.server_url,
        manifests_dir=manifests_dir,
        headers=config.user.auth_header,
        fides_key=fides_key,
        resource_type="data_use",
        all_resources_file=None,
    )


@pull.command(name="data_subject")  # type: ignore
@click.pass_context
@fides_key_argument
@manifests_dir_argument
def data_subject(
    ctx: click.Context,
    fides_key: str,
    manifests_dir: str,
) -> None:
    """
    Retrieve a specific data_subject from the server and update the local manifest files.
    """

    config = ctx.obj["CONFIG"]
    _pull.pull(
        url=config.cli.server_url,
        manifests_dir=manifests_dir,
        headers=config.user.auth_header,
        fides_key=fides_key,
        resource_type="data_subject",
        all_resources_file=None,
    )



from typing import List, Optional, TypeVar

from pydantic import BaseModel, ConfigDict, EmailStr, model_validator

from fides.api.schemas.base_class import NoValidationSchema


class IdentityTypes(BaseModel):
    email: bool
    phone_number: bool


class AdvancedSettings(BaseModel):
    identity_types: IdentityTypes


T = TypeVar("T")


class BaseEmailSchema(
    BaseModel,
):
    """
    Base Email Schema that contains the fields common to all email schemas.
    Used to validate secrets for all email connectors
    """

    test_email_address: Optional[EmailStr] = (
        None  # Email to send a connection test email
    )

    # the default value is temporary until we allow users to customize the identity types from the front-end
    advanced_settings: AdvancedSettings = AdvancedSettings(
        identity_types=IdentityTypes(email=True, phone_number=False)
    )
    model_config = ConfigDict(extra="forbid", from_attributes=True)

    @model_validator(mode="after")
    def validate_fields(self) -> "BaseEmailSchema":
        """At least one identity or browser identity needs to be specified on setup"""
        advanced_settings = self.advanced_settings
        if not advanced_settings:
            raise ValueError("Must supply advanced settings.")

        identities = advanced_settings.identity_types
        if not identities.email and not identities.phone_number:
            raise ValueError("Must supply at least one identity_type.")
        return self


class EmailSchema(BaseEmailSchema):
    """Schema to validate the secrets needed for a generic email connector"""

    third_party_vendor_name: str
    recipient_email_address: EmailStr


class EmailDocsSchema(EmailSchema, NoValidationSchema):
    """EmailDocsSchema Secrets Schema for API Docs"""


class ExtendedIdentityTypes(IdentityTypes):
    """Overrides basic IdentityTypes to add cookie_ids"""

    cookie_ids: List[str] = []


class AdvancedSettingsWithExtendedIdentityTypes(AdvancedSettings):
    """Overrides base AdvancedSettings to have extended IdentityTypes"""

    identity_types: ExtendedIdentityTypes


class ExtendedEmailSchema(EmailSchema):
    """Email schema used to unpack secrets for all email connector types (both generic, Sovrn, etc.)"""

    # the default value is temporary until we allow users to customize the identity types from the front-end
    advanced_settings: AdvancedSettingsWithExtendedIdentityTypes = (
        AdvancedSettingsWithExtendedIdentityTypes(
            identity_types=ExtendedIdentityTypes(
                email=True, phone_number=False, cookie_ids=[]
            )
        )
    )

    @model_validator(mode="after")
    def validate_fields(self) -> "ExtendedEmailSchema":
        """At least one identity or browser identity needs to be specified on setup"""
        advanced_settings = self.advanced_settings
        if not advanced_settings:
            raise ValueError("Must supply advanced settings.")

        identities = advanced_settings.identity_types
        if (
            not identities.email
            and not identities.phone_number
            and not identities.cookie_ids
        ):
            raise ValueError("Must supply at least one identity_type.")
        return self



"""update cookie max age to bigint

Revision ID: ddec29f24945
Revises: c5a218831820
Create Date: 2023-10-24 13:05:40.442124

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Integer
from sqlalchemy.dialects.postgresql import BIGINT

# revision identifiers, used by Alembic.
revision = "ddec29f24945"
down_revision = "c5a218831820"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("ctl_systems", "cookie_max_age_seconds", type_=BIGINT)


def downgrade():
    op.alter_column("ctl_systems", "cookie_max_age_seconds", type_=Integer)



from typing import Any, Dict, Optional

from boto3 import Session
from botocore.exceptions import ClientError
from loguru import logger

from fides.api.common_exceptions import StorageUploadError
from fides.api.schemas.storage.storage import AWSAuthMethod, StorageSecrets


def get_aws_session(
    auth_method: str,
    storage_secrets: Optional[Dict[StorageSecrets, Any]],
    assume_role_arn: Optional[str] = None,
) -> Session:
    """
    Abstraction to retrieve an AWS Session using secrets.

    If an `assume_role_arn` is provided, the secrets will be used to
    assume that role and return a Session instantiated with that role.
    """
    if storage_secrets is None:
        # set to an empty dict to allow for more dynamic code downstream
        storage_secrets = {}
    if auth_method == AWSAuthMethod.SECRET_KEYS.value:
        if not storage_secrets:
            err_msg = "Storage secrets not found for S3 storage."
            logger.warning(err_msg)
            raise StorageUploadError(err_msg)

        session = Session(
            aws_access_key_id=storage_secrets[StorageSecrets.AWS_ACCESS_KEY_ID.value],  # type: ignore
            aws_secret_access_key=storage_secrets[
                StorageSecrets.AWS_SECRET_ACCESS_KEY.value  # type: ignore
            ],
            region_name=storage_secrets.get("region_name"),  # type: ignore
        )
    elif auth_method == AWSAuthMethod.AUTOMATIC.value:
        session = Session(
            region_name=storage_secrets.get("region_name"),  # type: ignore
        )
        logger.info("Successfully created automatic session")
    else:
        logger.error("AWS auth method not supported: {}", auth_method)
        raise ValueError(f"AWS auth method not supported: {auth_method}")

    # Check that credentials are valid
    sts_client = session.client("sts")
    sts_client.get_caller_identity()

    if assume_role_arn:
        try:
            response = sts_client.assume_role(
                RoleArn=assume_role_arn, RoleSessionName="FidesAssumeRoleSession"
            )
            temp_credentials = response["Credentials"]
            logger.info(
                f"Assumed role {assume_role_arn} and got temporary credentials."
            )
            return Session(
                aws_access_key_id=temp_credentials["AccessKeyId"],
                aws_secret_access_key=temp_credentials["SecretAccessKey"],
                aws_session_token=temp_credentials["SessionToken"],
                region_name=storage_secrets.get("region_name"),  # type: ignore
            )
        except ClientError as error:
            logger.exception(
                f"Failed to assume assume role {assume_role_arn}. Error: {error.response['Error']['Message']}"
            )
            raise
    else:
        return session



"""remove_consent_settings

Revision ID: 9b98aba5bba8
Revises: 4cb3b5af4160
Create Date: 2023-10-03 20:26:12.492657

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "9b98aba5bba8"
down_revision = "4cb3b5af4160"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_consentsettings_id", table_name="consentsettings")
    op.drop_table("consentsettings")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "consentsettings",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=False),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("tcf_enabled", sa.BOOLEAN(), autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint("id", name="consentsettings_pkey"),
    )
    op.create_index("ix_consentsettings_id", "consentsettings", ["id"], unique=False)
    # ### end Alembic commands ###



"""Add privacy request status

Revision ID: d65e7e921814
Revises: c5336b841d70
Create Date: 2021-12-02 22:06:16.270442

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d65e7e921814"
down_revision = "c5336b841d70"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("alter type privacyrequeststatus add value 'paused'")


def downgrade():
    op.execute("delete from privacyrequest where status in ('paused')")
    op.execute("alter type privacyrequeststatus rename to privacyrequeststatus_old")
    op.execute(
        "create type privacyrequeststatus as enum('in_processing', 'complete', 'pending', 'error')"
    )
    op.execute(
        (
            "alter table privacyrequest alter column status type privacyrequeststatus using "
            "status::text::privacyrequeststatus"
        )
    )
    op.execute("drop type privacyrequeststatus_old")



from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

from fastapi.encoders import jsonable_encoder
from fideslang.models import Dataset as FideslangDataset
from loguru import logger
from pydantic import ValidationError as PydanticValidationError
from sqlalchemy.orm import Session

from fides.api.common_exceptions import (
    SaaSConfigNotFoundException,
    UnreachableNodesError,
    ValidationError,
)
from fides.api.graph.config import GraphDataset
from fides.api.graph.graph import DatasetGraph
from fides.api.graph.traversal import Traversal, TraversalNode
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.api import BulkUpdateFailed
from fides.api.schemas.dataset import (
    BulkPutDataset,
    DatasetConfigCtlDataset,
    ValidateDatasetResponse,
)
from fides.api.schemas.privacy_request import PrivacyRequestSource, PrivacyRequestStatus
from fides.api.schemas.redis_cache import Identity, LabeledIdentity
from fides.service.dataset.dataset_service import (
    DatasetNotFoundException,
    _get_ctl_dataset,
)
from fides.service.dataset.dataset_validator import DatasetValidator
from fides.service.dataset.validation_steps.traversal import TraversalValidationStep


class DatasetFilter:
    """
    Filter that excludes nodes that are not part of the specified dataset.
    This ensures that unreachable nodes from other datasets are not treated as errors.
    """

    def __init__(self, dataset_name: str):
        self.dataset_name = dataset_name

    def exclude_node(self, node: TraversalNode) -> bool:
        """Returns True if the node is not part of the dataset"""
        return node.address.dataset != self.dataset_name


class DatasetConfigService:
    def __init__(self, db: Session):
        self.db = db

    def create_or_update_dataset_config(
        self,
        connection_config: ConnectionConfig,
        dataset: Union[DatasetConfigCtlDataset, FideslangDataset],
    ) -> Tuple[Optional[FideslangDataset], Optional[BulkUpdateFailed]]:
        """Create or update a single dataset"""
        try:
            if isinstance(dataset, DatasetConfigCtlDataset):
                ctl_dataset = _get_ctl_dataset(self.db, dataset.ctl_dataset_fides_key)
                dataset_to_validate = FideslangDataset.model_validate(ctl_dataset)
                data_dict = {
                    "connection_config_id": connection_config.id,
                    "fides_key": dataset.fides_key,
                    "ctl_dataset_id": ctl_dataset.id,
                }
            else:
                dataset_to_validate = dataset
                data_dict = {
                    "connection_config_id": connection_config.id,
                    "fides_key": dataset.fides_key,
                    "dataset": dataset.model_dump(mode="json"),
                }

            # Validate dataset
            DatasetValidator(
                self.db,
                dataset_to_validate,
                connection_config,
                skip_steps=[TraversalValidationStep],
            ).validate()

            # Create or update using unified method
            dataset_config = DatasetConfig.create_or_update(self.db, data=data_dict)

            return dataset_config.ctl_dataset, None

        except (SaaSConfigNotFoundException, ValidationError) as exception:
            error = BulkUpdateFailed(
                message=str(exception),
                data=dataset.model_dump(),
            )
            logger.warning(f"Dataset validation failed: {str(exception)}")
            return None, error

        except (PydanticValidationError, DatasetNotFoundException):
            # Raising errors for now to stay consistent with existing behavior.
            # TODO: Include these errors in the bulk update response and update the test assertions.
            raise

        except Exception as e:
            message = f"Create/update failed for dataset '{dataset.fides_key}'"
            logger.warning(f"{message}: {str(e)}")
            error = BulkUpdateFailed(
                message="Dataset create/update failed.",
                data=dataset.model_dump(),
            )
            return None, error

    def bulk_create_or_update_dataset_configs(
        self,
        connection_config: ConnectionConfig,
        datasets: Union[List[DatasetConfigCtlDataset], List[FideslangDataset]],
    ) -> BulkPutDataset:
        """Create or update multiple datasets"""
        created_or_updated: List[FideslangDataset] = []
        failed: List[BulkUpdateFailed] = []

        logger.info("Starting bulk upsert for {} datasets", len(datasets))

        for item in datasets:
            dataset_result, error = self.create_or_update_dataset_config(
                connection_config=connection_config,
                dataset=item,
            )

            if dataset_result:
                created_or_updated.append(dataset_result)
            if error:
                failed.append(error)

        return BulkPutDataset(
            succeeded=created_or_updated,
            failed=failed,
        )

    def validate_dataset_config(
        self, connection_config: ConnectionConfig, dataset: FideslangDataset
    ) -> ValidateDatasetResponse:

        return DatasetValidator(self.db, dataset, connection_config).validate()

    def get_dataset_reachability(
        self, dataset_config: DatasetConfig, policy: Optional[Policy] = None
    ) -> Tuple[bool, Optional[Union[str, List[Dict[str, Any]]]]]:
        """
        Determines if the given dataset is reachable along with an error message
        """
        # First check if the target dataset is valid
        try:
            dataset_config.get_graph()
        except PydanticValidationError as exc:
            return False, jsonable_encoder(
                exc.errors(
                    include_url=False, include_input=False, include_context=False
                )
            )

        # Get graphs for all enabled datasets
        dataset_graphs = []
        datasets = DatasetConfig.all(db=self.db)
        for dataset in datasets:
            if not dataset.connection_config.disabled:
                try:
                    dataset_graphs.append(dataset.get_graph())
                except PydanticValidationError:
                    continue

        # We still want to check reachability even if our dataset config's connection is disabled.
        # We also consider the siblings, because if the connection is enabled, then all the
        # datasets will be enabled with it.
        sibling_dataset_graphs = []
        if dataset_config.connection_config.disabled:
            for sibling in dataset_config.connection_config.datasets:
                try:
                    sibling_dataset_graphs.append(sibling.get_graph())
                except PydanticValidationError:
                    continue

        try:
            dataset_graph = DatasetGraph(*dataset_graphs, *sibling_dataset_graphs)
        except ValidationError as exc:
            return (
                False,
                f'The following dataset references do not exist "{", ".join(exc.errors)}"',
            )

        # dummy data is enough to determine traversability
        identity_seed: Dict[str, str] = {
            k: "something" for k in dataset_graph.identity_keys.values()
        }

        try:
            Traversal(
                dataset_graph,
                identity_seed,
                policy,
                node_filters=[DatasetFilter(dataset_config.fides_key)],
            )
        except UnreachableNodesError as exc:
            return (
                False,
                f'The following collections are not reachable "{", ".join(exc.errors)}"',
            )

        return True, None

    def run_test_access_request(
        self,
        policy: Policy,
        dataset_config: DatasetConfig,
        input_data: Dict[str, Any],
    ) -> PrivacyRequest:
        """
        Creates a privacy request with a source of "Dataset test" that runs an access request for a single dataset.
        The input data is used to mock any external dataset values referenced by our dataset so that it can run in
        complete isolation.
        """

        # Create a privacy request with a source of "Dataset test"
        # so it's not shown to the user.
        privacy_request = PrivacyRequest.create(
            db=self.db,
            data={
                "policy_id": policy.id,
                "source": PrivacyRequestSource.dataset_test,
                "status": PrivacyRequestStatus.in_processing,
            },
        )

        with logger.contextualize(
            privacy_request_id=privacy_request.id,
            privacy_request_source=PrivacyRequestSource.dataset_test.value,
        ):
            try:
                # Remove periods and colons to avoid them being parsed as path delimiters downstream.
                escaped_input_data = {
                    key.replace(".", "_").replace(":", "_"): value
                    for key, value in input_data.items()
                }

                # Manually cache the input data as identity data.
                # We're doing a bit of trickery here to avoid asking for labels for custom identities.
                predefined_fields = Identity.model_fields.keys()
                input_identity = {
                    key: (
                        value
                        if key in predefined_fields
                        else LabeledIdentity(label=key, value=value)
                    )
                    for key, value in escaped_input_data.items()
                }
                privacy_request.cache_identity(input_identity)

                graph_dataset = dataset_config.get_graph()
                modified_graph_dataset = replace_references_with_identities(
                    dataset_config.fides_key, graph_dataset
                )

                dataset_graph = DatasetGraph(modified_graph_dataset)
                connection_config = dataset_config.connection_config

                from fides.api.task.create_request_tasks import run_access_request

                run_access_request(
                    privacy_request,
                    policy,
                    dataset_graph,
                    [connection_config],
                    escaped_input_data,
                    self.db,
                    privacy_request_proceed=False,
                )
            except Exception as exc:
                logger.error(f"Error running test access request: {exc}")

        return privacy_request


def replace_references_with_identities(
    dataset_key: str, graph_dataset: GraphDataset
) -> GraphDataset:
    """
    Replace external field references with identity values for testing.

    Creates a copy of the graph dataset and replaces dataset references with
    equivalent identity references that can be seeded directly. This allows
    testing a single dataset in isolation without needing to load data from
    referenced external datasets.
    """

    modified_graph_dataset = deepcopy(graph_dataset)

    for collection in modified_graph_dataset.collections:
        for field in collection.fields:
            for ref, edge_direction in field.references[:]:
                if edge_direction == "from" and ref.dataset != dataset_key:
                    field.identity = f"{ref.dataset}_{ref.collection}_{'_'.join(ref.field_path.levels)}"
                    field.references.remove((ref, "from"))

    return modified_graph_dataset



"""Contains all of the deploy group of CLI commands for fides."""

from os import environ
from subprocess import CalledProcessError
from typing import Optional

import rich_click as click

from fides.common.utils import echo_green, print_divider
from fides.config.create import create_and_update_config_file
from fides.core.deploy import (
    check_docker_version,
    check_fides_uploads_dir,
    check_virtualenv,
    print_deploy_success,
    pull_specific_docker_image,
    start_application,
    teardown_application,
)


# NOTE: This behaves similarly to 'init' in that it is excluded from analytics
# in the main CLI logic. This allows us to run `fides init` _during_ the deploy
# command, once the sample project server is accessible, to allow us to
# immediately register it
@click.group(name="deploy")
@click.pass_context
def deploy(ctx: click.Context) -> None:
    """
    Deploy a sample project locally to try out Fides.
    """


@deploy.command()  # type: ignore
@click.pass_context
@click.option(
    "--no-pull",
    is_flag=True,
    help="Use a local image instead of trying to pull from DockerHub.",
)
@click.option(
    "--no-init",
    is_flag=True,
    help="Disable the initialization of the Fides CLI, to run in headless mode.",
)
@click.option(
    "--env-file",
    type=click.Path(exists=True),
    help="Use a custom ENV file for the Fides container to override settings.",
)
@click.option(
    "--image",
    type=str,
    help="Use a custom image for the Fides container instead of the default ('ethyca/fides').",
)
def up(
    ctx: click.Context,
    no_pull: bool = False,
    no_init: bool = False,
    env_file: Optional[click.Path] = None,
    image: Optional[str] = None,
) -> None:  # pragma: no cover
    """
    Starts a sample project via docker compose.
    """

    check_virtualenv()
    try:
        check_docker_version()
    except:  # pylint: disable=bare-except
        response = click.confirm(
            "WARNING: Encountered an error while checking Docker versions. Would you like to attempt a deploy anyway?"
        )
        if not response:
            raise SystemExit("Deploy aborted!")
    else:
        echo_green("Docker version is compatible, starting deploy...")

    config = ctx.obj["CONFIG"]

    if not no_pull:
        pull_specific_docker_image()

    if env_file:
        print(f"> Using custom ENV file from: {env_file}")
        environ["FIDES_DEPLOY_ENV_FILE"] = str(env_file)

    if image:
        print(f"> Using custom image: {image}")
        environ["FIDES_DEPLOY_IMAGE"] = image

    try:
        check_fides_uploads_dir()
        print("> Starting application...")
        start_application()
        click.clear()

        # Deployment is ready! Perform the same steps as `fides init` to setup CLI
        if not no_init:
            echo_green("Deployment successful! Initializing fides...")
            create_and_update_config_file(config=config)
        else:
            echo_green(
                "Deployment successful! Skipping CLI initialization (run 'fides init' to initialize)"
            )
        print_divider()

        print_deploy_success()
    except CalledProcessError:
        teardown_application()
        raise SystemExit(1)


@deploy.command()
@click.pass_context
def down(ctx: click.Context) -> None:
    """
    Stops the sample project and removes all volumes.
    """

    check_docker_version()
    teardown_application()



"""split privacy experience table

Revision ID: b546ce845e6c
Revises: fc04e3e637c0
Create Date: 2023-05-11 16:21:13.825649

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "b546ce845e6c"
down_revision = "fc04e3e637c0"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("delete from privacyexperiencehistory;")
    op.execute("delete from privacyexperience;")
    op.create_table(
        "privacyexperienceconfig",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("acknowledgement_button_label", sa.String(), nullable=True),
        sa.Column("banner_title", sa.String(), nullable=True),
        sa.Column("banner_description", sa.String(), nullable=True),
        sa.Column("component", sa.String(), nullable=False),
        sa.Column("component_title", sa.String(), nullable=True),
        sa.Column("component_description", sa.String(), nullable=True),
        sa.Column("confirmation_button_label", sa.String(), nullable=True),
        sa.Column("delivery_mechanism", sa.String(), nullable=False),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("is_default", sa.Boolean(), nullable=False),
        sa.Column("link_label", sa.String(), nullable=True),
        sa.Column("reject_button_label", sa.String(), nullable=True),
        sa.Column("version", sa.Float(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyexperienceconfig_component"),
        "privacyexperienceconfig",
        ["component"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperienceconfig_delivery_mechanism"),
        "privacyexperienceconfig",
        ["delivery_mechanism"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperienceconfig_id"),
        "privacyexperienceconfig",
        ["id"],
        unique=False,
    )
    op.create_table(
        "privacyexperienceconfighistory",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("acknowledgement_button_label", sa.String(), nullable=True),
        sa.Column("banner_title", sa.String(), nullable=True),
        sa.Column("banner_description", sa.String(), nullable=True),
        sa.Column("component", sa.String(), nullable=False),
        sa.Column("component_title", sa.String(), nullable=True),
        sa.Column("component_description", sa.String(), nullable=True),
        sa.Column("confirmation_button_label", sa.String(), nullable=True),
        sa.Column("delivery_mechanism", sa.String(), nullable=False),
        sa.Column("disabled", sa.Boolean(), nullable=False),
        sa.Column("is_default", sa.Boolean(), nullable=False),
        sa.Column("link_label", sa.String(), nullable=True),
        sa.Column("reject_button_label", sa.String(), nullable=True),
        sa.Column("version", sa.Float(), nullable=False),
        sa.Column("experience_config_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["experience_config_id"],
            ["privacyexperienceconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyexperienceconfighistory_component"),
        "privacyexperienceconfighistory",
        ["component"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperienceconfighistory_delivery_mechanism"),
        "privacyexperienceconfighistory",
        ["delivery_mechanism"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperienceconfighistory_id"),
        "privacyexperienceconfighistory",
        ["id"],
        unique=False,
    )
    op.drop_index(
        "ix_privacyexperiencetemplate_id", table_name="privacyexperiencetemplate"
    )
    op.drop_index(
        "ix_privacyexperiencetemplate_regions", table_name="privacyexperiencetemplate"
    )
    op.add_column("privacyexperience", sa.Column("region", sa.String(), nullable=False))
    op.add_column(
        "privacyexperience",
        sa.Column("experience_config_id", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperience",
        sa.Column("experience_config_history_id", sa.String(), nullable=True),
    )
    op.drop_index("ix_privacyexperience_regions", table_name="privacyexperience")
    op.create_index(
        op.f("ix_privacyexperience_experience_config_history_id"),
        "privacyexperience",
        ["experience_config_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperience_experience_config_id"),
        "privacyexperience",
        ["experience_config_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperience_region"),
        "privacyexperience",
        ["region"],
        unique=False,
    )
    op.create_unique_constraint(
        "region_component", "privacyexperience", ["region", "component"]
    )
    op.drop_constraint(
        "privacyexperience_privacy_experience_template_id_fkey",
        "privacyexperience",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "privacyexperience_experience_config_history_id_fkey",
        "privacyexperience",
        "privacyexperienceconfighistory",
        ["experience_config_history_id"],
        ["id"],
    )
    op.create_foreign_key(
        "privacyexperience_experience_config_id_fkey",
        "privacyexperience",
        "privacyexperienceconfig",
        ["experience_config_id"],
        ["id"],
    )
    op.drop_column("privacyexperience", "banner_description")
    op.drop_column("privacyexperience", "confirmation_button_label")
    op.drop_column("privacyexperience", "component_description")
    op.drop_column("privacyexperience", "banner_title")
    op.drop_column("privacyexperience", "acknowledgement_button_label")
    op.drop_column("privacyexperience", "reject_button_label")
    op.drop_column("privacyexperience", "component_title")
    op.drop_column("privacyexperience", "privacy_experience_template_id")
    op.drop_column("privacyexperience", "regions")
    op.drop_column("privacyexperience", "link_label")
    op.add_column(
        "privacyexperiencehistory", sa.Column("region", sa.String(), nullable=False)
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column("experience_config_id", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column("experience_config_history_id", sa.String(), nullable=True),
    )
    op.drop_index(
        "ix_privacyexperiencehistory_regions", table_name="privacyexperiencehistory"
    )
    op.create_index(
        op.f("ix_privacyexperiencehistory_experience_config_history_id"),
        "privacyexperiencehistory",
        ["experience_config_history_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperiencehistory_experience_config_id"),
        "privacyexperiencehistory",
        ["experience_config_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperiencehistory_privacy_experience_id"),
        "privacyexperiencehistory",
        ["privacy_experience_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyexperiencehistory_region"),
        "privacyexperiencehistory",
        ["region"],
        unique=False,
    )
    op.drop_constraint(
        "privacyexperiencehistory_privacy_experience_template_id_fkey",
        "privacyexperiencehistory",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "privacyexperiencehistory_experience_config_id_fkey",
        "privacyexperiencehistory",
        "privacyexperienceconfig",
        ["experience_config_id"],
        ["id"],
    )
    op.create_foreign_key(
        "privacyexperiencehistory_experience_config_history_id_fkey",
        "privacyexperiencehistory",
        "privacyexperienceconfighistory",
        ["experience_config_history_id"],
        ["id"],
    )
    op.drop_column("privacyexperiencehistory", "banner_description")
    op.drop_column("privacyexperiencehistory", "confirmation_button_label")
    op.drop_column("privacyexperiencehistory", "component_description")
    op.drop_column("privacyexperiencehistory", "banner_title")
    op.drop_column("privacyexperiencehistory", "acknowledgement_button_label")
    op.drop_column("privacyexperiencehistory", "reject_button_label")
    op.drop_column("privacyexperiencehistory", "component_title")
    op.drop_column("privacyexperiencehistory", "privacy_experience_template_id")
    op.drop_column("privacyexperiencehistory", "regions")
    op.drop_column("privacyexperiencehistory", "link_label")
    op.drop_table("privacyexperiencetemplate")


def downgrade():
    op.add_column(
        "privacyexperiencehistory",
        sa.Column("link_label", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "regions",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "privacy_experience_template_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column("component_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "reject_button_label", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "acknowledgement_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column("banner_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "component_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "confirmation_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "banner_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.drop_constraint(
        "privacyexperiencehistory_experience_config_id_fkey",
        "privacyexperiencehistory",
        type_="foreignkey",
    )
    op.drop_constraint(
        "privacyexperiencehistory_experience_config_history_id_fkey",
        "privacyexperiencehistory",
        type_="foreignkey",
    )

    op.drop_index(
        op.f("ix_privacyexperiencehistory_region"),
        table_name="privacyexperiencehistory",
    )
    op.drop_index(
        op.f("ix_privacyexperiencehistory_privacy_experience_id"),
        table_name="privacyexperiencehistory",
    )
    op.drop_index(
        op.f("ix_privacyexperiencehistory_experience_config_id"),
        table_name="privacyexperiencehistory",
    )
    op.drop_index(
        op.f("ix_privacyexperiencehistory_experience_config_history_id"),
        table_name="privacyexperiencehistory",
    )
    op.create_index(
        "ix_privacyexperiencehistory_regions",
        "privacyexperiencehistory",
        ["regions"],
        unique=False,
    )
    op.drop_column("privacyexperiencehistory", "experience_config_history_id")
    op.drop_column("privacyexperiencehistory", "experience_config_id")
    op.drop_column("privacyexperiencehistory", "region")
    op.add_column(
        "privacyexperience",
        sa.Column("link_label", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "regions",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "privacy_experience_template_id",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column("component_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "reject_button_label", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "acknowledgement_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column("banner_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "component_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "confirmation_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperience",
        sa.Column(
            "banner_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.drop_constraint(
        "privacyexperience_experience_config_id_fkey",
        "privacyexperience",
        type_="foreignkey",
    )
    op.drop_constraint(
        "privacyexperience_experience_config_history_id_fkey",
        "privacyexperience",
        type_="foreignkey",
    )

    op.drop_constraint("region_component", "privacyexperience", type_="unique")
    op.drop_index(op.f("ix_privacyexperience_region"), table_name="privacyexperience")
    op.drop_index(
        op.f("ix_privacyexperience_experience_config_id"),
        table_name="privacyexperience",
    )
    op.drop_index(
        op.f("ix_privacyexperience_experience_config_history_id"),
        table_name="privacyexperience",
    )
    op.create_index(
        "ix_privacyexperience_regions", "privacyexperience", ["regions"], unique=False
    )
    op.drop_column("privacyexperience", "experience_config_history_id")
    op.drop_column("privacyexperience", "experience_config_id")
    op.drop_column("privacyexperience", "region")
    op.create_table(
        "privacyexperiencetemplate",
        sa.Column("id", sa.VARCHAR(length=255), autoincrement=False, nullable=False),
        sa.Column(
            "created_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            postgresql.TIMESTAMP(timezone=True),
            server_default=sa.text("now()"),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("disabled", sa.BOOLEAN(), autoincrement=False, nullable=False),
        sa.Column("component", sa.VARCHAR(), autoincrement=False, nullable=False),
        sa.Column(
            "delivery_mechanism", sa.VARCHAR(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "regions",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("component_title", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "component_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column("banner_title", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "banner_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column("link_label", sa.VARCHAR(), autoincrement=False, nullable=True),
        sa.Column(
            "confirmation_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column(
            "reject_button_label", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "acknowledgement_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
        sa.PrimaryKeyConstraint("id", name="privacyexperiencetemplate_pkey"),
    )
    op.create_index(
        "ix_privacyexperiencetemplate_regions",
        "privacyexperiencetemplate",
        ["regions"],
        unique=False,
    )
    op.create_index(
        "ix_privacyexperiencetemplate_id",
        "privacyexperiencetemplate",
        ["id"],
        unique=False,
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfighistory_id"),
        table_name="privacyexperienceconfighistory",
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfighistory_delivery_mechanism"),
        table_name="privacyexperienceconfighistory",
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfighistory_component"),
        table_name="privacyexperienceconfighistory",
    )
    op.drop_table("privacyexperienceconfighistory")
    op.drop_index(
        op.f("ix_privacyexperienceconfig_id"), table_name="privacyexperienceconfig"
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfig_delivery_mechanism"),
        table_name="privacyexperienceconfig",
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfig_component"),
        table_name="privacyexperienceconfig",
    )
    op.drop_table("privacyexperienceconfig")
    op.create_foreign_key(
        "privacyexperiencehistory_privacy_experience_template_id_fkey",
        "privacyexperiencehistory",
        "privacyexperiencetemplate",
        ["privacy_experience_template_id"],
        ["id"],
    )
    op.create_foreign_key(
        "privacyexperience_privacy_experience_template_id_fkey",
        "privacyexperience",
        "privacyexperiencetemplate",
        ["privacy_experience_template_id"],
        ["id"],
    )



import itertools
from collections import defaultdict
from typing import Any, Dict, List, Optional, Set, Union

from loguru import logger

from fides.api.graph.config import CollectionAddress, FieldPath
from fides.api.graph.graph import DatasetGraph
from fides.api.util.collection_util import Row


def filter_data_categories(
    access_request_results: Dict[str, List[Dict[str, Optional[Any]]]],
    target_categories: Set[str],
    dataset_graph: DatasetGraph,
    rule_key: str = "",
    fides_connector_datasets: Optional[Set[str]] = None,
) -> Dict[str, List[Dict[str, Optional[Any]]]]:
    """Filter access request results to only return fields associated with the target data categories
    and subcategories.

    Regarding subcategories, if data category "user.contact" is specified on one of the rule
    targets, for example, all fields on subcategories also apply, so ["user.contact.address.city",
    "user.contact.address.street", ...], etc.

    :param access_request_results: Dictionary of access request results for each of your collections
    :param target_categories: A set of data categories that we'd like to extract from access_request_results
    :param dataset_graph: The dataset graph where the data category mappings are accessed from

    :return: Filtered access request results that only contain fields matching the desired data categories.
    """
    logger.info(
        "Filtering Access Request results to return fields associated with data categories"
    )
    filtered_access_results: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
    for node_address, results in access_request_results.items():
        if not results:
            continue

        # Results from fides connectors are a special case:
        # they've already been filtered and stored in a dict keyed by rule key.
        # So here, we simply find the results corresponding to our current rule
        # and unpack the result so that its stored at the "top level"
        # of the results dict
        if fides_connector_datasets and (
            CollectionAddress.from_string(node_address).dataset
            in fides_connector_datasets
        ):
            unpack_fides_connector_results(
                results, filtered_access_results, rule_key, node_address
            )
            # do not do any further processing on fides connector results
            # as they have already been pre-filtered
            continue

        # Gets all FieldPaths on this traversal_node associated with the requested data
        # categories and sub data categories
        target_field_paths: Set[FieldPath] = set(
            itertools.chain(
                *[
                    field_paths
                    for cat, field_paths in dataset_graph.data_category_field_mapping[
                        CollectionAddress.from_string(node_address)
                    ].items()
                    if any(cat.startswith(tar) for tar in target_categories)
                ]
            )
        )

        collection_data_categories = set(
            dataset_graph.nodes[
                CollectionAddress.from_string(node_address)
            ].collection.data_categories
            or []
        )

        if collection_data_categories:
            if any(
                collection_category.startswith(tuple(target_categories))
                for collection_category in collection_data_categories
            ):
                filtered_access_results[node_address].extend(results)
                continue

        if not target_field_paths:
            continue

        for row in results:
            filtered_results: Dict[str, Any] = {}
            for field_path in target_field_paths:
                select_and_save_field(filtered_results, row, field_path)
            remove_empty_containers(filtered_results)
            filtered_access_results[node_address].append(filtered_results)

    return filtered_access_results


def select_and_save_field(saved: Any, row: Row, target_path: FieldPath) -> Dict:
    """Extract the data located along the given `target_path` from the row and add to the "saved" dictionary.

    Entire rows are returned from your collections; this function will incrementally just pull the PII from the rows,
    by retrieving data along target_paths to relevant data categories.

    To use, pass in an empty dict for "saved" and loop through a list of FieldPaths you want,
    continuing to pass in the ever-growing new "saved" dict that was returned from the previous iteration.

    :param saved: Call with an empty dict to start, it will recursively update as data along the target_path is added.
    :param row: Call with retrieved row to start, it will recursively be called with a variety of object types until we
    reach the most deeply nested value.
    :param target_path: FieldPath to the data we want to retrieve

    :return: modified saved dictionary with given field path if found
    """

    def _defaultdict_or_array(resource: Any) -> Any:
        """Helper for building new nested resource - can return an empty dict, empty array or resource itself"""
        return type(resource)() if isinstance(resource, (list, dict)) else resource

    if isinstance(row, list):
        for i, elem in enumerate(row):
            try:
                saved[i] = select_and_save_field(saved[i], elem, target_path)
            except IndexError:
                saved.append(
                    select_and_save_field(
                        _defaultdict_or_array(elem), elem, target_path
                    )
                )

    elif isinstance(row, dict):
        for key in row:
            if target_path.levels and key == target_path.levels[0]:
                if key not in saved:
                    saved[key] = _defaultdict_or_array(row[key])
                saved[key] = select_and_save_field(
                    saved[key], row[key], FieldPath(*target_path.levels[1:])
                )
    return saved


RecursiveRow = Union[Dict[Any, Any], List[Any]]


def remove_empty_containers(row: RecursiveRow) -> RecursiveRow:
    """
    Recursively updates row in place to remove empty dictionaries and empty arrays at any level in collection or
    from embedded collections in arrays.

    `select_and_save_field` recursively builds a nested structure based on desired field paths.
    If no input data was found along a deeply nested field path, we may have empty dicts to clean up
    before supplying response to user.  Also empty arrays and empty dicts do not contain PII.

    :param row: Pass in retrieved row, and it will recursively go through objects and arrays and filter out empty collections.
    :return: Updated row with empty objects and arrays removed
    """
    if isinstance(row, dict):
        for key, value in row.copy().items():
            if isinstance(value, (dict, list)):
                value = remove_empty_containers(value)

            if value in [{}, []]:
                del row[key]

    elif isinstance(row, list):
        for index, elem in reversed(list(enumerate(row))):
            if isinstance(elem, (dict, list)):
                elem = remove_empty_containers(elem)

            if elem in [{}, []]:
                row.pop(index)

    return row


def unpack_fides_connector_results(
    connector_results: List[Dict[str, Any]],
    filtered_access_results: Dict[str, List[Dict[str, Any]]],
    rule_key: str,
    node_address: str,
) -> None:
    """
    Unpacks the pre-filtered results from a Fides connector
    that are associated with the given `rule_key`.
    These results are stored in the provided aggregate `filtered_acces_results` dict.
    """
    # there should only ever be one element in the list here, since the
    # "real" data is nested one level deeper. but we will iterate through
    # anyway, just to be safe.
    for results in connector_results:
        try:
            rule_results = results[rule_key]
        except KeyError:
            logger.error(
                "Did not find a result entry on Fides connector {} for rule {}",
                node_address,
                rule_key,
            )
            continue

        try:
            filtered_access_results.update(rule_results)
        except ValueError:
            # if we get a ValueError, a given CollectionAddress key
            # already exists in the results dict. Handle that here.
            for key, value in rule_results.items():
                filtered: Optional[List[Dict[str, Optional[Any]]]] = (
                    filtered_access_results.get(key)
                )
                if not filtered:
                    filtered_access_results[key] = value  # type: ignore
                else:
                    if value:
                        logger.info("Appending child rows to {}", key)
                        filtered.extend(value)  # type: ignore



"""new tables for pre-approval webhooks

Revision ID: 6cfd59e7920a
Revises: d49a767eb49d
Create Date: 2024-04-15 13:14:39.027682

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6cfd59e7920a"
down_revision = "d49a767eb49d"
branch_labels = None
depends_on = None


def upgrade():
    # Add new table: PreApprovalWebhook
    op.create_table(
        "preapprovalwebhook",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["connection_config_id"],
            ["connectionconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_preapprovalwebhook_id"), "preapprovalwebhook", ["id"], unique=False
    )

    # Add new table: PreApprovalWebhookReply
    op.create_table(
        "preapprovalwebhookreply",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column("webhook_id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("privacy_request_id", sa.String(), nullable=True),
        sa.Column("is_eligible", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["privacy_request_id"],
            ["privacyrequest.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_preapprovalwebhookreply_id"),
        "preapprovalwebhookreply",
        ["id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # Remove PreApprovalWebhook and PreApprovalWebhookReply tables
    op.drop_index(op.f("ix_preapprovalwebhook_id"), table_name="preapprovalwebhook")
    op.drop_index(
        op.f("ix_preapprovalwebhookreply_id"), table_name="preapprovalwebhookreply"
    )
    op.drop_table("preapprovalwebhook")
    op.drop_table("preapprovalwebhookreply")
    # ### end Alembic commands ###



from typing import Any, Dict, List, Optional, Tuple, Union

from fastapi import HTTPException
from sqlalchemy.orm import Query, Session
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.privacy_notice import (
    EnforcementLevel,
    PrivacyNotice,
    UserConsentPreference,
)
from fides.api.models.privacy_preference import PrivacyPreferenceHistory
from fides.api.models.privacy_request import (
    PrivacyRequest,
    ProvidedIdentity,
    ProvidedIdentityType,
)
from fides.api.models.sql_models import System  # type: ignore[attr-defined]
from fides.api.models.tcf_purpose_overrides import TCFPurposeOverride
from fides.api.schemas.privacy_request import ExecutionLogStatus
from fides.api.schemas.redis_cache import Identity


def filter_privacy_preferences_for_propagation(
    system: Optional[System],
    privacy_preferences: List[PrivacyPreferenceHistory],
) -> List[PrivacyPreferenceHistory]:
    """Filter privacy preferences on a privacy request to just the ones that should be considered for third party
    consent propagation.

    Only applies to preferences saved for privacy notices here, not against individual TCF components.
    """

    propagatable_preferences: List[PrivacyPreferenceHistory] = [
        pref
        for pref in privacy_preferences
        if pref.privacy_notice_history
        and pref.privacy_notice_history.enforcement_level
        == EnforcementLevel.system_wide
        and (
            pref.preference
            and pref.preference
            in [UserConsentPreference.opt_in, UserConsentPreference.opt_out]
        )
    ]

    if not system:
        return propagatable_preferences

    filtered_on_use: List[PrivacyPreferenceHistory] = []
    for pref in propagatable_preferences:
        if (
            pref.privacy_notice_history
            and pref.privacy_notice_history.applies_to_system(system)
        ):
            filtered_on_use.append(pref)
    return filtered_on_use


def build_user_consent_and_filtered_preferences_for_service(
    system: Optional[System],
    privacy_request: PrivacyRequest,
    session: Session,
    is_notice_based: bool = False,
) -> Tuple[
    Optional[Union[bool, Dict[str, UserConsentPreference]]],
    List[PrivacyPreferenceHistory],
]:
    """
    For SaaS Connectors, examine the Privacy Preferences and collapse this information into a single should we opt in? (True),
    should we opt out? (False) or should we do nothing? (None).

    Email connectors should instead call "filter_privacy_preferences_for_propagation" directly, since we can have preferences with
    conflicting opt in/opt out values for those connector types.

    Also return filtered preferences here so we can cache affected systems and/or secondary identifiers directly on these
    filtered preferences for consent reporting.

    - If using the old workflow (privacyrequest.consent_preferences), return True if all attached consent preferences
    are opt in, otherwise False.  System check is ignored.

    - If using the new workflow (privacyrequest.privacy_preferences), there is more filtering here.  Privacy Preferences
    must have an enforcement level of system-wide and a data use must match a system data use.  If the connector is
    orphaned (no system), skip the data use check. If conflicts, prefer the opt-out preference.
    """

    # OLD WORKFLOW
    if privacy_request.consent_preferences:
        return (
            all(
                consent_pref["opt_in"]
                for consent_pref in privacy_request.consent_preferences
            ),
            [],  # Don't need to return the filtered preferences, this is just relevant for the new workflow
        )

    # NEW WORKFLOWS: 1.notice-based and 2. global SaaS Consent
    relevant_preferences: List[PrivacyPreferenceHistory] = (
        filter_privacy_preferences_for_propagation(
            system,
            privacy_request.privacy_preferences,  # type: ignore[attr-defined]
        )
    )
    if not relevant_preferences:
        return None, []  # We should do nothing here

    filtered_preferences: List[PrivacyPreferenceHistory] = []

    # 1. NOTICE-BASED WORKFLOW
    if is_notice_based:
        notice_id_to_preference_map: Dict[str, UserConsentPreference] = {}
        # retrieve notices from the DB if they are associated with a relevant preference
        notices_with_preference: Query = session.query(PrivacyNotice).filter(
            PrivacyNotice.notice_key.in_(
                [preference.notice_key for preference in relevant_preferences]
            ),
        )
        # build our notice id -> user preference map, build filtered list of preferences that apply
        for notice in notices_with_preference:
            preference = [
                pref
                for pref in relevant_preferences
                if pref.notice_key == notice.notice_key
            ]
            # we only expect max 1 preference in this list
            if preference and preference[0]:
                notice_id_to_preference_map[notice.id] = preference[0].preference  # type: ignore[assignment]
                filtered_preferences.append(preference[0])

        return notice_id_to_preference_map, filtered_preferences

    # 2. GLOBAL (OPT-IN/OUT) WORKFLOW
    preference_to_propagate: UserConsentPreference = (
        UserConsentPreference.opt_out
        if any(
            filtered_pref.preference == UserConsentPreference.opt_out
            for filtered_pref in relevant_preferences
        )
        else UserConsentPreference.opt_in
    )

    # Hopefully rare final filtering in case there are conflicting preferences
    filtered_preferences = [
        pref
        for pref in relevant_preferences
        if pref.preference == preference_to_propagate
    ]

    # Return whether we should opt in, and the filtered preferences so we can update those for consent reporting
    return (
        preference_to_propagate == UserConsentPreference.opt_in,
        filtered_preferences,
    )


def cache_initial_status_and_identities_for_consent_reporting(
    db: Session,
    privacy_request: PrivacyRequest,
    connection_config: ConnectionConfig,
    relevant_preferences: List[PrivacyPreferenceHistory],
    relevant_user_identities: Dict[str, Any],
) -> None:
    """Add a pending system status and cache relevant identities on the applicable PrivacyPreferenceHistory
    records for consent reporting.

    Preferences that aren't relevant for the given system/connector are given a skipped status.

    Typically used when *some* but not all privacy preferences are relevant.  Otherwise,
    other methods just mark all the preferences as skipped.
    """
    for pref in privacy_request.privacy_preferences:  # type: ignore[attr-defined]
        if pref in relevant_preferences:
            pref.update_secondary_user_ids(db, relevant_user_identities)
            pref.cache_system_status(
                db, connection_config.system_key, ExecutionLogStatus.pending
            )
        else:
            pref.cache_system_status(
                db, connection_config.system_key, ExecutionLogStatus.skipped
            )


def add_complete_system_status_for_consent_reporting(
    db: Session,
    privacy_request: PrivacyRequest,
    connection_config: ConnectionConfig,
) -> None:
    """Cache a complete system status for consent reporting on just the subset
    of preferences that were deemed relevant for the connector on failure

    Deeming them relevant if they already had a "pending" log added to them.
    """
    for pref in privacy_request.privacy_preferences:  # type: ignore[attr-defined]
        if (
            pref.affected_system_status
            and pref.affected_system_status.get(connection_config.system_key)
            == ExecutionLogStatus.pending.value
        ):
            pref.cache_system_status(
                db,
                connection_config.system_key,
                ExecutionLogStatus.complete,
            )


def add_errored_system_status_for_consent_reporting(
    db: Session,
    privacy_request: PrivacyRequest,
    connection_config: ConnectionConfig,
) -> None:
    """Cache an errored system status for consent reporting on just the subset
    of preferences that were deemed relevant for the connector on failure

    Deeming them relevant if they already had a "pending" log added to them.
    """
    for pref in privacy_request.privacy_preferences:  # type: ignore[attr-defined]
        if (
            pref.affected_system_status
            and pref.affected_system_status.get(connection_config.system_key)
            == ExecutionLogStatus.pending.value
        ):
            pref.cache_system_status(
                db,
                connection_config.system_key,
                ExecutionLogStatus.error,
            )


def get_fides_user_device_id_provided_identity(
    db: Session, fides_user_device_id: Optional[str]
) -> Optional[ProvidedIdentity]:
    """Look up a fides user device id that is not attached to a privacy request if it exists

    There can be many fides user device ids attached to privacy requests, but we should try to keep them
    unique for consent requests.
    """
    if not fides_user_device_id:
        return None

    return ProvidedIdentity.filter(
        db=db,
        conditions=(
            (
                ProvidedIdentity.field_name
                == ProvidedIdentityType.fides_user_device_id.value
            )
            & (
                ProvidedIdentity.hashed_value.in_(
                    ProvidedIdentity.hash_value_for_search(fides_user_device_id)
                )
            )
            & (ProvidedIdentity.privacy_request_id.is_(None))
        ),
    ).first()


def get_or_create_fides_user_device_id_provided_identity(
    db: Session,
    identity_data: Optional[Identity],
) -> ProvidedIdentity:
    """Gets an existing fides user device id provided identity or creates one if it doesn't exist.
    Raises an error if no fides user device id is supplied.
    """
    if not identity_data or not identity_data.fides_user_device_id:
        raise HTTPException(
            HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Fides user device id not found in identity data",
        )

    identity = get_fides_user_device_id_provided_identity(
        db, identity_data.fides_user_device_id
    )

    if not identity:
        identity = ProvidedIdentity.create(
            db,
            data={
                "privacy_request_id": None,
                "field_name": ProvidedIdentityType.fides_user_device_id.value,
                "hashed_value": ProvidedIdentity.hash_value(
                    identity_data.fides_user_device_id
                ),
                "encrypted_value": {"value": identity_data.fides_user_device_id},
            },
        )

    return identity  # type: ignore[return-value]


def create_default_tcf_purpose_overrides_on_startup(
    db: Session,
) -> List[TCFPurposeOverride]:
    """On startup, load default Purpose Overrides, one for each purpose, with a default of is_included=True
    and no legal basis override

    The defaults have no effect on what is returned in the TCF Privacy Experience, and this functionality needs
    to be enabled via a config variable to be used at all.
    """
    purpose_override_resources_created: List[TCFPurposeOverride] = []

    for purpose_id in range(1, 12):
        if (
            not db.query(TCFPurposeOverride)
            .filter(TCFPurposeOverride.purpose == purpose_id)
            .first()
        ):
            purpose_override_resources_created.append(
                TCFPurposeOverride.create(
                    db, data={"purpose": purpose_id, "is_included": True}
                )
            )

    return purpose_override_resources_created



"""system_dictionary_migration

Revision ID: 39397820a0d5
Revises: 3a47ce736a37
Create Date: 2023-08-03 20:07:46.326248

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "39397820a0d5"
down_revision = "3a47ce736a37"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("ctl_systems", sa.Column("vendor_id", sa.String(), nullable=True))
    op.add_column(
        "ctl_systems",
        sa.Column(
            "dataset_references",
            sa.ARRAY(sa.String()),
            server_default="{}",
            nullable=False,
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "processes_personal_data", sa.BOOLEAN(), server_default="t", nullable=False
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "exempt_from_privacy_regulations",
            sa.BOOLEAN(),
            server_default="f",
            nullable=False,
        ),
    )
    op.add_column(
        "ctl_systems", sa.Column("reason_for_exemption", sa.String(), nullable=True)
    )
    op.add_column(
        "ctl_systems",
        sa.Column("uses_profiling", sa.BOOLEAN(), server_default="f", nullable=False),
    )
    op.add_column(
        "ctl_systems",
        sa.Column("legal_basis_for_profiling", sa.String(), nullable=True),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "does_international_transfers",
            sa.BOOLEAN(),
            server_default="f",
            nullable=False,
        ),
    )
    op.add_column(
        "ctl_systems",
        sa.Column("legal_basis_for_transfers", sa.String(), nullable=True),
    )
    op.add_column(
        "ctl_systems",
        sa.Column(
            "requires_data_protection_assessments",
            sa.BOOLEAN(),
            server_default="f",
            nullable=False,
        ),
    )
    op.add_column("ctl_systems", sa.Column("dpa_location", sa.String(), nullable=True))
    op.add_column("ctl_systems", sa.Column("dpa_progress", sa.String(), nullable=True))

    op.add_column(
        "ctl_systems", sa.Column("privacy_policy", sa.String(), nullable=True)
    )
    op.add_column("ctl_systems", sa.Column("legal_name", sa.String(), nullable=True))
    op.add_column("ctl_systems", sa.Column("legal_address", sa.String(), nullable=True))
    op.add_column(
        "ctl_systems",
        sa.Column(
            "responsibility", sa.ARRAY(sa.String()), server_default="{}", nullable=True
        ),
    )
    op.add_column("ctl_systems", sa.Column("dpo", sa.String(), nullable=True))
    op.add_column(
        "ctl_systems", sa.Column("joint_controller_info", sa.String(), nullable=True)
    )
    op.add_column(
        "ctl_systems", sa.Column("data_security_practices", sa.String(), nullable=True)
    )
    op.add_column(
        "privacydeclaration",
        sa.Column(
            "features", sa.ARRAY(sa.String()), server_default="{}", nullable=False
        ),
    )
    op.add_column(
        "privacydeclaration",
        sa.Column("legal_basis_for_processing", sa.String(), nullable=True),
    )
    op.add_column(
        "privacydeclaration",
        sa.Column("impact_assessment_location", sa.String(), nullable=True),
    )
    op.add_column(
        "privacydeclaration", sa.Column("retention_period", sa.String(), nullable=True)
    )
    op.add_column(
        "privacydeclaration",
        sa.Column(
            "processes_special_category_data",
            sa.BOOLEAN(),
            server_default="f",
            nullable=False,
        ),
    )
    op.add_column(
        "privacydeclaration",
        sa.Column("special_category_legal_basis", sa.String(), nullable=True),
    )
    op.add_column(
        "privacydeclaration",
        sa.Column(
            "data_shared_with_third_parties",
            sa.BOOLEAN(),
            server_default="f",
            nullable=False,
        ),
    )
    op.add_column(
        "privacydeclaration", sa.Column("third_parties", sa.String(), nullable=True)
    )
    op.add_column(
        "privacydeclaration",
        sa.Column(
            "shared_categories",
            sa.ARRAY(sa.String()),
            server_default="{}",
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacydeclaration", "shared_categories")
    op.drop_column("privacydeclaration", "third_parties")
    op.drop_column("privacydeclaration", "data_shared_with_third_parties")
    op.drop_column("privacydeclaration", "special_category_legal_basis")
    op.drop_column("privacydeclaration", "processes_special_category_data")
    op.drop_column("privacydeclaration", "retention_period")
    op.drop_column("privacydeclaration", "impact_assessment_location")
    op.drop_column("privacydeclaration", "legal_basis_for_processing")
    op.drop_column("privacydeclaration", "features")
    op.drop_column("ctl_systems", "data_security_practices")
    op.drop_column("ctl_systems", "joint_controller_info")
    op.drop_column("ctl_systems", "dpo")
    op.drop_column("ctl_systems", "responsibility")
    op.drop_column("ctl_systems", "legal_address")
    op.drop_column("ctl_systems", "legal_name")
    op.drop_column("ctl_systems", "privacy_policy")
    op.drop_column("ctl_systems", "dpa_progress")
    op.drop_column("ctl_systems", "dpa_location")
    op.drop_column("ctl_systems", "requires_data_protection_assessments")
    op.drop_column("ctl_systems", "legal_basis_for_transfers")
    op.drop_column("ctl_systems", "does_international_transfers")
    op.drop_column("ctl_systems", "legal_basis_for_profiling")
    op.drop_column("ctl_systems", "uses_profiling")
    op.drop_column("ctl_systems", "reason_for_exemption")
    op.drop_column("ctl_systems", "exempt_from_privacy_regulations")
    op.drop_column("ctl_systems", "processes_personal_data")
    op.drop_column("ctl_systems", "dataset_references")
    op.drop_column("ctl_systems", "vendor_id")
    # ### end Alembic commands ###
